"""StableDiffusionXL Module
"""
import gc
import sys
import os

sys.path.insert(0, os.path.dirname(__file__))

from ssl import Options
from nova_utils.interfaces.server_module import Processor
from diffusers import StableDiffusionXLImg2ImgPipeline, StableDiffusionXLPipeline, logging
from compel import Compel, ReturnedEmbeddingsType
from nova_utils.utils.cache_utils import get_file
import numpy as np
PYTORCH_ENABLE_MPS_FALLBACK = 1

import torch
from PIL import Image
from lora import build_lora_xl
logging.disable_progress_bar()
logging.enable_explicit_format()
#logging.set_verbosity_info()


# Setting defaults
_default_options = {"model": "stabilityai/stable-diffusion-xl-base-1.0", "ratio": "1-1", "width": "", "height":"", "high_noise_frac" : "0.8", "n_steps" : "35", "lora" : "" }

# TODO: add log infos, 
class StableDiffusionXL(Processor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.options = _default_options | self.options
        self.device = None
        self.ds_iter = None
        self.current_session = None
       

        # IO shortcuts
        self.input = [x for x in self.model_io if x.io_type == "input"]
        self.output = [x for x in self.model_io if x.io_type == "output"]
        self.input = self.input[0]
        self.output = self.output[0]

    def process_data(self, ds_iter) -> dict:
        self._device = ("cuda" if torch.cuda.is_available() else ("mps" if torch.backends.mps.is_built() else "cpu"))
        self.variant = "fp16"
        self.torch_d_type = torch.float16
        self.ds_iter = ds_iter
        current_session_name = self.ds_iter.session_names[0]
        self.current_session = self.ds_iter.sessions[current_session_name]['manager']      
        input_prompt = self.current_session.input_data['input_prompt'].data
        input_prompt = ' '.join(input_prompt)
        negative_prompt = self.current_session.input_data['negative_prompt'].data
        negative_prompt = ' '.join(negative_prompt)
        new_width = 0
        new_height = 0
        print("Input prompt: " + input_prompt)
        print("Negative prompt: " + negative_prompt)

        try:
            if self.options['width'] != "" and self.options['height'] != "":
                new_width = int(self.options['width'])
                new_height = int(self.options['height'])                        
                ratiow, ratioh =  self.calculate_aspect(new_width, new_height) 
                print("Ratio:" + str(ratiow) + ":" + str(ratioh))

            else:
                ratiow = str(self.options['ratio']).split('-')[0]
                ratioh =str(self.options['ratio']).split('-')[1]

            model = self.options["model"]
            lora = self.options["lora"]
            mwidth = 1024
            mheight = 1024

            height = mheight
            width = mwidth

            ratiown = int(ratiow)
            ratiohn= int(ratioh)

            if ratiown > ratiohn:
                height = int((ratiohn/ratiown) * float(width))
            elif ratiown < ratiohn:
                width = int((ratiown/ratiohn) * float(height))
            elif ratiown == ratiohn:
                width = height

        
            print("Processing Output width: " + str(width) + " Output height: " + str(height))

          


            if model == "stabilityai/stable-diffusion-xl-base-1.0":
                base = StableDiffusionXLPipeline.from_pretrained(model, torch_dtype=self.torch_d_type, variant=self.variant, use_safetensors=True).to(self.device)
                print("Loaded model: " + model)

            else:
                 
                model_uri = [ x for x in self.trainer.meta_uri if x.uri_id == model][0]
                if str(model_uri) == "":
                    return  "Model not found"

                model_path = get_file(
                    fname=str(model_uri.uri_id) + ".safetensors",
                    origin=model_uri.uri_url,
                    file_hash=model_uri.uri_hash,
                    cache_dir=os.getenv("CACHE_DIR"),
                    tmp_dir=os.getenv("TMP_DIR"),
                    )
                
                print(str(model_path))


                base = StableDiffusionXLPipeline.from_single_file(str(model_path), torch_dtype=self.torch_d_type, variant=self.variant, use_safetensors=True).to(self.device)
                print("Loaded model: " + model)

            if lora != "" and lora != "None":
                print("Loading lora...")          
                lora, input_prompt, existing_lora =  build_lora_xl(lora, input_prompt, "")

                if existing_lora:
                    lora_uri = [ x for x in self.trainer.meta_uri if x.uri_id == lora][0]
                    if str(lora_uri) == "":
                        return  "Lora not found"
                    lora_path = get_file(
                        fname=str(lora_uri.uri_id) + ".safetensors",
                        origin=lora_uri.uri_url,
                        file_hash=lora_uri.uri_hash,
                        cache_dir=os.getenv("CACHE_DIR"),
                        tmp_dir=os.getenv("TMP_DIR"),
                        )
                    
                    base.load_lora_weights(str(lora_path))
                    print("Loaded Lora: " + str(lora_path))

            refiner = StableDiffusionXLImg2ImgPipeline.from_pretrained(
            "stabilityai/stable-diffusion-xl-refiner-1.0",
            text_encoder_2=base.text_encoder_2,
            vae=base.vae,
            torch_dtype=self.torch_d_type,
            use_safetensors=True,
            variant=self.variant,
            )

            
            compel_base = Compel(
                tokenizer=[base.tokenizer, base.tokenizer_2],
                text_encoder=[base.text_encoder, base.text_encoder_2],
                returned_embeddings_type=ReturnedEmbeddingsType.PENULTIMATE_HIDDEN_STATES_NON_NORMALIZED,
                requires_pooled=[False, True],
            )

            compel_refiner = Compel(
                tokenizer=[refiner.tokenizer_2],
                text_encoder=[refiner.text_encoder_2],
                returned_embeddings_type=ReturnedEmbeddingsType.PENULTIMATE_HIDDEN_STATES_NON_NORMALIZED,
                requires_pooled=[True])

            conditioning, pooled = compel_base(input_prompt)
            negative_conditioning, negative_pooled = compel_base(negative_prompt)

            conditioning_refiner, pooled_refiner = compel_refiner(input_prompt)
            negative_conditioning_refiner, negative_pooled_refiner = compel_refiner(
                negative_prompt)
        
            
            n_steps = int(self.options['n_steps'])
            high_noise_frac = float(self.options['high_noise_frac'])


            #base.unet = torch.compile(base.unet, mode="reduce-overhead", fullgraph=True)



            img = base(
                prompt_embeds=conditioning,
                pooled_prompt_embeds=pooled,
                negative_prompt_embeds=negative_conditioning,
                negative_pooled_prompt_embeds=negative_pooled,
                width=width,
                height=height,
                num_inference_steps=n_steps,
                denoising_end=high_noise_frac,
                output_type="latent",
            ).images

            if torch.cuda.is_available():
                del base
                gc.collect()
                torch.cuda.empty_cache()
                torch.cuda.ipc_collect()

            refiner.to(self.device)
            # refiner.enable_model_cpu_offload()
            image = refiner(
                prompt_embeds=conditioning_refiner,
                pooled_prompt_embeds=pooled_refiner,
                negative_prompt_embeds=negative_conditioning_refiner,
                negative_pooled_prompt_embeds=negative_pooled_refiner,
                num_inference_steps=n_steps,
                denoising_start=high_noise_frac,
                num_images_per_prompt=1,
                image=img,
            ).images[0]

            if torch.cuda.is_available():
                del refiner
                gc.collect()
                torch.cuda.empty_cache()
                torch.cuda.ipc_collect()

            if new_height != 0 or new_width != 0 and (new_width != mwidth or new_height != mheight) :
                print("Resizing to width: " + str(new_width) + " height: " + str(new_height))
                image = image.resize((new_width, new_height), Image.LANCZOS)
    
            numpy_array = np.array(image)
            return numpy_array


        except Exception as e:
            print(e)
            sys.stdout.flush()
            return "Error"
        
    def calculate_aspect(self, width: int, height: int):
        def gcd(a, b):
            """The GCD (greatest common divisor) is the highest number that evenly divides both width and height."""
            return a if b == 0 else gcd(b, a % b)

        r = gcd(width, height)
        x = int(width / r)
        y = int(height / r)

        return x, y


    
    def to_output(self, data: dict):
        self.current_session.output_data_templates['output_image'].data = data
        return self.current_session.output_data_templates


import json
import os
import re
import sys
import urllib.parse
from typing import Any
from urllib.request import urlopen, Request

import requests
import instaloader
from pytube import YouTube


def XitterDownload(source_url, target_location):
    script_dir = os.path.dirname(os.path.realpath(__file__))
    request_details_file = f"{script_dir}{os.sep}request_details.json"
    request_details = json.load(open(request_details_file, "r"))  # test
    features, variables = request_details["features"], request_details["variables"]

    def get_tokens(tweet_url):
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:84.0) Gecko/20100101 Firefox/84.0",
            "Accept": "*/*",
            "Accept-Language": "de,en-US;q=0.7,en;q=0.3",
            "Accept-Encoding": "gzip, deflate, br",
            "TE": "trailers",
        }

        html = requests.get(tweet_url, headers=headers)

        assert (
                html.status_code == 200
        ), f"Failed to get tweet page.  If you are using the correct Twitter URL this suggests a bug in the script.  Please open a GitHub issue and copy and paste this message.  Status code: {html.status_code}.  Tweet url: {tweet_url}"

        mainjs_url = re.findall(
            r"https://abs.twimg.com/responsive-web/client-web-legacy/main.[^\.]+.js",
            html.text,
        )

        assert (
                mainjs_url is not None and len(mainjs_url) > 0
        ), f"Failed to find main.js file.  If you are using the correct Twitter URL this suggests a bug in the script.  Please open a GitHub issue and copy and paste this message.  Tweet url: {tweet_url}"

        mainjs_url = mainjs_url[0]
        mainjs = requests.get(mainjs_url)

        assert (
                mainjs.status_code == 200
        ), f"Failed to get main.js file.  If you are using the correct Twitter URL this suggests a bug in the script.  Please open a GitHub issue and copy and paste this message.  Status code: {mainjs.status_code}.  Tweet url: {tweet_url}"

        bearer_token = re.findall(r'AAAAAAAAA[^"]+', mainjs.text)

        assert (
                bearer_token is not None and len(bearer_token) > 0
        ), f"Failed to find bearer token.  If you are using the correct Twitter URL this suggests a bug in the script.  Please open a GitHub issue and copy and paste this message.  Tweet url: {tweet_url}, main.js url: {mainjs_url}"

        bearer_token = bearer_token[0]

        # get the guest token
        with requests.Session() as s:
            s.headers.update(
                {
                    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:84.0) Gecko/20100101 Firefox/84.0",
                    "accept": "*/*",
                    "accept-language": "de,en-US;q=0.7,en;q=0.3",
                    "accept-encoding": "gzip, deflate, br",
                    "te": "trailers",
                }
            )

            s.headers.update({"authorization": f"Bearer {bearer_token}"})

            # activate bearer token and get guest token
            guest_token = s.post("https://api.twitter.com/1.1/guest/activate.json").json()[
                "guest_token"
            ]

        assert (
                guest_token is not None
        ), f"Failed to find guest token.  If you are using the correct Twitter URL this suggests a bug in the script.  Please open a GitHub issue and copy and paste this message.  Tweet url: {tweet_url}, main.js url: {mainjs_url}"

        return bearer_token, guest_token

    def get_details_url(tweet_id, features, variables):
        # create a copy of variables - we don't want to modify the original
        variables = {**variables}
        variables["tweetId"] = tweet_id

        return f"https://twitter.com/i/api/graphql/0hWvDhmW8YQ-S_ib3azIrw/TweetResultByRestId?variables={urllib.parse.quote(json.dumps(variables))}&features={urllib.parse.quote(json.dumps(features))}"
        # return f"https://api.twitter.com/graphql/ncDeACNGIApPMaqGVuF_rw/TweetResultByRestId?variables={urllib.parse.quote(json.dumps(variables))}&features={urllib.parse.quote(json.dumps(features))}"

    def get_tweet_details(tweet_url, guest_token, bearer_token):
        tweet_id = re.findall(r"(?<=status/)\d+", tweet_url)

        assert (
                tweet_id is not None and len(tweet_id) == 1
        ), f"Could not parse tweet id from your url.  Make sure you are using the correct url.  If you are, then file a GitHub issue and copy and paste this message.  Tweet url: {tweet_url}"

        tweet_id = tweet_id[0]

        # the url needs a url encoded version of variables and features as a query string
        url = get_details_url(tweet_id, features, variables)

        details = requests.get(
            url,
            headers={
                "authorization": f"Bearer {bearer_token}",
                "x-guest-token": guest_token,
            },
        )

        max_retries = 10
        cur_retry = 0
        while details.status_code == 400 and cur_retry < max_retries:
            try:
                error_json = json.loads(details.text)
            except json.JSONDecodeError:
                assert (
                    False
                ), f"Failed to parse json from details error. details text: {details.text}  If you are using the correct Twitter URL this suggests a bug in the script.  Please open a GitHub issue and copy and paste this message.  Status code: {details.status_code}.  Tweet url: {tweet_url}"

            assert (
                    "errors" in error_json
            ), f"Failed to find errors in details error json.  If you are using the correct Twitter URL this suggests a bug in the script.  Please open a GitHub issue and copy and paste this message.  Status code: {details.status_code}.  Tweet url: {tweet_url}"

            needed_variable_pattern = re.compile(r"Variable '([^']+)'")
            needed_features_pattern = re.compile(
                r'The following features cannot be null: ([^"]+)'
            )

            for error in error_json["errors"]:
                needed_vars = needed_variable_pattern.findall(error["message"])
                for needed_var in needed_vars:
                    variables[needed_var] = True

                needed_features = needed_features_pattern.findall(error["message"])
                for nf in needed_features:
                    for feature in nf.split(","):
                        features[feature.strip()] = True

            url = get_details_url(tweet_id, features, variables)

            details = requests.get(
                url,
                headers={
                    "authorization": f"Bearer {bearer_token}",
                    "x-guest-token": guest_token,
                },
            )

            cur_retry += 1

            if details.status_code == 200:
                # save new variables
                request_details["variables"] = variables
                request_details["features"] = features

                with open(request_details_file, "w") as f:
                    json.dump(request_details, f, indent=4)

        assert (
                details.status_code == 200
        ), f"Failed to get tweet details.  If you are using the correct Twitter URL this suggests a bug in the script.  Please open a GitHub issue and copy and paste this message.  Status code: {details.status_code}.  Tweet url: {tweet_url}"

        return details

    def get_tweet_status_id(tweet_url):
        sid_patern = r'https://(?:x\.com|twitter\.com)/[^/]+/status/(\d+)'
        if tweet_url[len(tweet_url) - 1] != "/":
            tweet_url = tweet_url + "/"

        match = re.findall(sid_patern, tweet_url)
        if len(match) == 0:
            print("error, could not get status id from this tweet url :", tweet_url)
            exit()
        status_id = match[0]
        return status_id

    def get_associated_media_id(j, tweet_url):
        sid = get_tweet_status_id(tweet_url)
        pattern = (
                r'"expanded_url"\s*:\s*"https://x\.com/[^/]+/status/'
                + sid
                + r'/[^"]+",\s*"id_str"\s*:\s*"\d+",'
        )
        matches = re.findall(pattern, j)
        if len(matches) > 0:
            target = matches[0]
            target = target[0: len(target) - 1]  # remove the coma at the end
            return json.loads("{" + target + "}")["id_str"]
        return None

    def extract_mp4s(j, tweet_url, target_all_mp4s=False):
        # pattern looks like https://video.twimg.com/amplify_video/1638969830442237953/vid/1080x1920/lXSFa54mAVp7KHim.mp4?tag=16 or https://video.twimg.com/ext_tw_video/1451958820348080133/pu/vid/720x1280/GddnMJ7KszCQQFvA.mp4?tag=12
        amplitude_pattern = re.compile(
            r"(https://video.twimg.com/amplify_video/(\d+)/vid/(\d+x\d+)/[^.]+.mp4\?tag=\d+)"
        )
        ext_tw_pattern = re.compile(
            r"(https://video.twimg.com/ext_tw_video/(\d+)/pu/vid/(avc1/)?(\d+x\d+)/[^.]+.mp4\?tag=\d+)"
        )
        # format - https://video.twimg.com/tweet_video/Fvh6brqWAAQhU9p.mp4
        tweet_video_pattern = re.compile(r'https://video.twimg.com/tweet_video/[^"]+')

        # https://video.twimg.com/ext_tw_video/1451958820348080133/pu/pl/b-CiC-gZClIwXgDz.m3u8?tag=12&container=fmp4
        container_pattern = re.compile(r'https://video.twimg.com/[^"]*container=fmp4')
        media_id = get_associated_media_id(j, tweet_url)
        # find all the matches
        matches = amplitude_pattern.findall(j)
        matches += ext_tw_pattern.findall(j)
        container_matches = container_pattern.findall(j)

        tweet_video_matches = tweet_video_pattern.findall(j)

        if len(matches) == 0 and len(tweet_video_matches) > 0:
            return tweet_video_matches

        results = {}

        for match in matches:
            url, tweet_id, _, resolution = match
            if tweet_id not in results:
                results[tweet_id] = {"resolution": resolution, "url": url}
            else:
                # if we already have a higher resolution video, then don't overwrite it
                my_dims = [int(x) for x in resolution.split("x")]
                their_dims = [int(x) for x in results[tweet_id]["resolution"].split("x")]

                if my_dims[0] * my_dims[1] > their_dims[0] * their_dims[1]:
                    results[tweet_id] = {"resolution": resolution, "url": url}

        if media_id:
            all_urls = []
            for twid in results:
                all_urls.append(results[twid]["url"])
            all_urls += container_matches

            url_with_media_id = []
            for url in all_urls:
                if url.__contains__(media_id):
                    url_with_media_id.append(url)

            if len(url_with_media_id) > 0:
                return url_with_media_id

        if len(container_matches) > 0 and not target_all_mp4s:
            return container_matches

        if target_all_mp4s:
            urls = [x["url"] for x in results.values()]
            urls += container_matches
            return urls
        return [x["url"] for x in results.values()]

    def extract_mp4_fmp4(j):
        """
        Extract the URL of the MP4 video from the detailed information of the tweet.
        Returns a list of URLs, tweet IDs, and resolution information (dictionary type)
        and a list of tweet IDs as return values.
        """

        # Empty list to store tweet IDs
        tweet_id_list = []
        mp4_info_dict_list = []
        amplitude_pattern = re.compile(
            r"(https://video.twimg.com/amplify_video/(\d+)/vid/(avc1/)(\d+x\d+)/[^.]+.mp4\?tag=\d+)"
        )
        ext_tw_pattern = re.compile(
            r"(https://video.twimg.com/ext_tw_video/(\d+)/pu/vid/(avc1/)?(\d+x\d+)/[^.]+.mp4\?tag=\d+)"
        )
        tweet_video_pattern = re.compile(r'https://video.twimg.com/tweet_video/[^"]+')
        container_pattern = re.compile(r'https://video.twimg.com/[^"]*container=fmp4')

        matches = amplitude_pattern.findall(j)
        matches += ext_tw_pattern.findall(j)
        container_matches = container_pattern.findall(j)
        tweet_video_url_list = tweet_video_pattern.findall(j)

        for match in matches:
            url, tweet_id, _, resolution = match
            tweet_id_list.append(int(tweet_id))
            mp4_info_dict_list.append({"resolution": resolution, "url": url})

        tweet_id_list = list(dict.fromkeys(tweet_id_list))

        if len(container_matches) > 0:
            for url in container_matches:
                mp4_info_dict_list.append({"url": url})

        return tweet_id_list, mp4_info_dict_list, tweet_video_url_list

    def download_parts(url, output_filename):
        resp = requests.get(url, stream=True)
        pattern = re.compile(r"(/[^\n]*/(\d+x\d+)/[^\n]*container=fmp4)")
        matches = pattern.findall(resp.text)
        max_res = 0
        max_res_url = None

        for match in matches:
            url, resolution = match
            width, height = resolution.split("x")
            res = int(width) * int(height)
            if res > max_res:
                max_res = res
                max_res_url = url

        assert (
                max_res_url is not None
        ), f"Could not find a url to download from.  Make sure you are using the correct url.  If you are, then file a GitHub issue and copy and paste this message.  Tweet url: {url}"

        video_part_prefix = "https://video.twimg.com"

        resp = requests.get(video_part_prefix + max_res_url, stream=True)

        mp4_pattern = re.compile(r"(/[^\n]*\.mp4)")
        mp4_parts = mp4_pattern.findall(resp.text)

        assert (
                len(mp4_parts) == 1
        ), f"There should be exactly 1 mp4 container at this point.  Instead, found {len(mp4_parts)}.  Please open a GitHub issue and copy and paste this message into it.  Tweet url: {url}"

        mp4_url = video_part_prefix + mp4_parts[0]

        m4s_part_pattern = re.compile(r"(/[^\n]*\.m4s)")
        m4s_parts = m4s_part_pattern.findall(resp.text)

        with open(output_filename, "wb") as f:
            r = requests.get(mp4_url, stream=True)
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
                    f.flush()

            for part in m4s_parts:
                part_url = video_part_prefix + part
                r = requests.get(part_url, stream=True)
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
                        f.flush()

        return True

    def repost_check(j, exclude_replies=True):
        try:
            reply_index = j.index('"conversationthread-')
        except ValueError:
            reply_index = len(j)
        if exclude_replies:
            j = j[0:reply_index]

        # We use this regular expression to extract the source status
        source_status_pattern = r'"source_status_id_str"\s*:\s*"\d+"'
        matches = re.findall(source_status_pattern, j)

        if len(matches) > 0 and exclude_replies:
            # We extract the source status id (ssid)
            ssid = json.loads("{" + matches[0] + "}")["source_status_id_str"]
            # We plug it in this regular expression to find expanded_url (the original tweet url)
            expanded_url_pattern = (
                    r'"expanded_url"\s*:\s*"https://x\.com/[^/]+/status/' + ssid + '[^"]+"'
            )
            matches2 = re.findall(expanded_url_pattern, j)

            if len(matches2) > 0:
                # We extract the url and return it
                status_url = json.loads("{" + matches2[0] + "}")["expanded_url"]
                return status_url

        if not exclude_replies:
            # If we include replies we'll have to get all ssids and remove duplicates
            ssids = []
            for match in matches:
                ssids.append(json.loads("{" + match + "}")["source_status_id_str"])
            # we remove duplicates (this line is messy but it's the easiest way to do it)
            ssids = list(set(ssids))
            if len(ssids) > 0:
                for ssid in ssids:
                    expanded_url_pattern = (
                            r'"expanded_url"\s*:\s*"https://x\.com/[^/]+/status/'
                            + ssid
                            + '[^"]+"'
                    )
                    matches2 = re.findall(expanded_url_pattern, j)
                    if len(matches2) > 0:
                        status_urls = []
                        for match in matches2:
                            status_urls.append(
                                json.loads("{" + match + "}")["expanded_url"]
                            )
                        # We remove duplicates another time
                        status_urls = list(set(status_urls))
                        return status_urls

        # If we don't find source_status_id_str, the tweet doesn't feature a reposted video
        return None

    def download_video_from_x(tweet_url, output_file, target_all_videos=False):
        bearer_token, guest_token = get_tokens(tweet_url)
        resp = get_tweet_details(tweet_url, guest_token, bearer_token)
        mp4s = extract_mp4s(resp.text, tweet_url, target_all_videos)

        if target_all_videos:
            video_counter = 1
            original_urls = repost_check(resp.text, exclude_replies=False)

            if len(original_urls) > 0:
                for url in original_urls:
                    download_video_from_x(
                        url, output_file.replace(".mp4", f"_{video_counter}.mp4")
                    )
                    video_counter += 1
                if len(mp4s) > 0:
                    for mp4 in mp4s:
                        output_file = output_file.replace(".mp4", f"_{video_counter}.mp4")
                        if "container" in mp4:
                            download_parts(mp4, output_file)

                        else:
                            # use a stream to download the file
                            r = requests.get(mp4, stream=True)
                            with open(output_file, "wb") as f:
                                for chunk in r.iter_content(chunk_size=1024):
                                    if chunk:
                                        f.write(chunk)
                                        f.flush()
                        video_counter += 1
        else:
            original_url = repost_check(resp.text)

            if original_url:
                download_video_from_x(original_url, output_file)
            else:
                assert (
                        len(mp4s) > 0
                ), f"Could not find any mp4s to download.  Make sure you are using the correct url.  If you are, then file a GitHub issue and copy and paste this message.  Tweet url: {tweet_url}"

                mp4 = mp4s[0]
                if "container" in mp4:
                    download_parts(mp4, output_file)
                else:
                    # use a stream to download the file
                    r = requests.get(mp4, stream=True)
                    with open(output_file, "wb") as f:
                        for chunk in r.iter_content(chunk_size=1024):
                            if chunk:
                                f.write(chunk)
                                f.flush()
        return target_location

    return download_video_from_x(source_url, target_location)


# TIKTOK/INSTA
def getDict() -> dict:
    response = requests.get('https://ttdownloader.com/')
    point = response.text.find('<input type="hidden" id="token" name="token" value="') + \
            len('<input type="hidden" id="token" name="token" value="')
    token = response.text[point:point + 64]
    TTDict = {
        'token': token,
    }

    for i in response.cookies:
        TTDict[str(i).split()[1].split('=')[0].strip()] = str(
            i).split()[1].split('=')[1].strip()
    return TTDict


def createHeader(parseDict) -> tuple[dict[str, Any], dict[str | Any, str | Any], dict[str, str | Any]]:
    cookies = {
        'PHPSESSID': parseDict['PHPSESSID'],
        # 'popCookie': parseDict['popCookie'],
    }
    headers = {
        'authority': 'ttdownloader.com',
        'accept': '*/*',
        'accept-language': 'en-US,en;q=0.9',
        'content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
        'origin': 'https://ttdownloader.com',
        'referer': 'https://ttdownloader.com/',
        'sec-ch-ua': '"Not?A_Brand";v="8", "Chromium";v="108", "Google Chrome";v="108"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-origin',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
                      'Chrome/108.0.0.0 Safari/537.36',
        'x-requested-with': 'XMLHttpRequest',
    }
    data = {
        'url': '',
        'format': '',
        'token': parseDict['token'],
    }
    return cookies, headers, data


def TikTokDownload(cookies, headers, data, name, path) -> str:
    response = requests.post('https://ttdownloader.com/search/',
                             cookies=cookies, headers=headers, data=data)
    parsed_link = [i for i in str(response.text).split()
                   if i.startswith("href=")][0]

    response = requests.get(parsed_link[6:-10])
    with open(path + "\\" + name + ".mp4", "wb") as f:
        f.write(response.content)
    return path + "\\" + name + ".mp4"


def TiktokDownloadAll(linkList, path) -> str:
    parseDict = getDict()
    cookies, headers, data = createHeader(parseDict)
    # linkList = getLinkDict()['tiktok']
    for i in linkList:
        try:
            data['url'] = i
            result = TikTokDownload(cookies, headers, data, "tiktok", path)  # str(linkList.index(i))
            return result
        except IndexError:
            parseDict = getDict()
            cookies, headers, data = createHeader(parseDict)
        except Exception as err:
            print(err)
            exit(1)


def InstagramDownload(url, name, path) -> str:
    obj = instaloader.Instaloader()
    post = instaloader.Post.from_shortcode(obj.context, url.split("/")[-2])
    photo_url = post.url
    video_url = post.video_url
    print(video_url)
    if video_url:
        response = requests.get(video_url)
        with open(path + "\\" + name + ".mp4", "wb") as f:
            f.write(response.content)
            return path + "\\" + name + ".mp4"
    elif photo_url:
        response = requests.get(photo_url)
        with open(path + "\\" + name + ".jpg", "wb") as f:
            f.write(response.content)
            return path + "\\" + name + ".jpg"


def InstagramDownloadAll(linklist, path) -> str:
    for i in linklist:
        try:
            print(str(linklist.index(i)))
            print(str(linklist[i]))
            result = InstagramDownload(i, str(linklist.index(i)), path)
            return result
        except Exception as err:
            print(err)
            exit(1)


# YOUTUBE
def YouTubeDownload(link, path, audio_only=True):
    youtubeObject = YouTube(link)
    if audio_only:
        youtubeObject = youtubeObject.streams.get_audio_only()
        youtubeObject.download(path, "yt.mp3")
        print("Download is completed successfully")
        return path + "yt.mp3"
    else:
        youtubeObject = youtubeObject.streams.get_highest_resolution()
        youtubeObject.download(path, "yt.mp4")
        print("Download is completed successfully")
        return path + "yt.mp4"


def checkYoutubeLinkValid(link):
    try:
        # TODO find a way to test without fully downloading the file
        youtubeObject = YouTube(link)
        youtubeObject = youtubeObject.streams.get_audio_only()
        youtubeObject.download(".", "yt.mp3")
        os.remove("yt.mp3")
        return True

    except Exception as e:
        print(str(e))
        return False


# OVERCAST
def OvercastDownload(source_url, target_location):
    def get_title(html_str):
        """Get the title from the meta tags"""

        title = re.findall(r"<meta name=\"og:title\" content=\"(.+)\"", html_str)
        if len(title) == 1:
            return title[0].replace("&mdash;", "-")
        return None

    def get_description(html_str):
        """Get the description from the Meta tag"""

        desc_re = r"<meta name=\"og:description\" content=\"(.+)\""
        description = re.findall(desc_re, html_str)
        if len(description) == 1:
            return description[0]
        return None

    def get_url(html_string):
        """Find the URL from the <audio><source>.... tag"""

        url = re.findall(r"<source src=\"(.+?)\"", html_string)
        if len(url) == 1:
            # strip off the last 4 characters to cater for the #t=0 in the URL
            # which urlretrieve flags as invalid
            return url[0][:-4]
        return None

    """Given a Overcast source URL fetch the file it points to"""
    headers = {
        "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) "
                      "AppleWebKit/537.11 (KHTML, like Gecko) "
                      "Chrome/23.0.1271.64 Safari/537.11",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
        "Accept-Charset": "ISO-8859-1,utf-8;q=0.7,*;q=0.3",
        "Accept-Encoding": "none",
        "Accept-Language": "en-US,en;q=0.8",
        "Connection": "keep-alive",
    }
    req = Request(source_url, None, headers)
    source_data = urlopen(req).read().decode('utf-8')
    title = get_title(source_data)
    url = get_url(source_data)

    if url is None or title is None:
        sys.exit("Could not find parse URL")
    if not os.path.exists(target_location):
        req = requests.get(url)
        file = open(target_location, 'wb')
        for chunk in req.iter_content(100000):
            file.write(chunk)
        file.close()



import json
import os
from datetime import timedelta
from nostr_sdk import Client, Timestamp, PublicKey, Tag, Keys, Options, SecretKey, NostrSigner, NostrDatabase, \
    ClientBuilder, Filter, NegentropyOptions, NegentropyDirection, init_logger, LogLevel, Kind

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag
from nostr_dvm.utils.output_utils import post_process_list_to_events, post_process_list_to_users

"""
This File contains a Module to search for notes
Accepted Inputs: a search query
Outputs: A list of events 
Params:  None
"""


class DiscoveryBotFarms(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_PEOPLE_DISCOVERY
    TASK: str = "discover-bot-farms"
    FIX_COST: float = 0
    dvm_config: DVMConfig
    last_schedule: int = 0

    def __init__(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                 admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)
        super().__init__(name=name, dvm_config=dvm_config, nip89config=nip89config, nip88config=nip88config,
                         admin_config=admin_config, options=options)

        use_logger = False
        if use_logger:
            init_logger(LogLevel.DEBUG)

        self.sync_db()

    def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "text":
                    return False
        return True

    def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        self.dvm_config = dvm_config
        print(self.dvm_config.PRIVATE_KEY)

        request_form = {"jobID": event.id().to_hex()}

        # default values
        search = "airdrop;just your average nostr enjoyer" #;@nostrich.house;
        max_results = 500

        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                if input_type == "text":
                    search = tag.as_vec()[1]
            elif tag.as_vec()[0] == 'param':
                param = tag.as_vec()[1]
                if param == "max_results":  # check for param type
                    max_results = int(tag.as_vec()[2])

        options = {
            "search": search,
            "max_results": max_results,
        }
        request_form['options'] = json.dumps(options)
        return request_form

    def process(self, request_form):
        from nostr_sdk import Filter
        options = self.set_options(request_form)

        opts = (Options().wait_for_send(False).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT)))
        sk = SecretKey.from_hex(self.dvm_config.PRIVATE_KEY)
        keys = Keys.parse(sk.to_hex())
        signer = NostrSigner.keys(keys)

        database = NostrDatabase.sqlite("db/nostr_profiles.db")
        cli = ClientBuilder().database(database).signer(signer).opts(opts).build()

        cli.add_relay("wss://relay.damus.io")
        # cli.add_relay("wss://atl.purplerelay.com")
        cli.connect()

        # Negentropy reconciliation

        # Query events from database

        filter1 = Filter().kind(Kind(0))
        events = cli.database().query([filter1])
        # for event in events:
        #    print(event.as_json())

        # events = cli.get_events_of([notes_filter], timedelta(seconds=5))

        result_list = []
        print("Events: " + str(len(events)))

        searchterms = str(options["search"]).split(";")
        index = 0
        if len(events) > 0:

            for event in events:
                if index < options["max_results"]:
                    try:
                        if any(ext in event.content().lower() for ext in searchterms):
                            p_tag = Tag.parse(["p", event.author().to_hex()])
                            print(event.as_json())
                            result_list.append(p_tag.as_vec())
                            index += 1
                    except Exception as exp:
                        print(str(exp) + " " + event.author().to_hex())
                else:
                    break

        return json.dumps(result_list)

    def post_process(self, result, event):
        """Overwrite the interface function to return a social client readable format, if requested"""
        for tag in event.tags():
            if tag.as_vec()[0] == 'output':
                format = tag.as_vec()[1]
                if format == "text/plain":  # check for output type
                    result = post_process_list_to_users(result)

        # if not text/plain, don't post-process
        return result

    def schedule(self, dvm_config):
        if dvm_config.SCHEDULE_UPDATES_SECONDS == 0:
            return 0
        else:
            if Timestamp.now().as_secs() >= self.last_schedule + dvm_config.SCHEDULE_UPDATES_SECONDS:
                self.sync_db()
                self.last_schedule = Timestamp.now().as_secs()
                return 1

    def sync_db(self):
        opts = (Options().wait_for_send(False).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT)))
        sk = SecretKey.from_hex(self.dvm_config.PRIVATE_KEY)
        keys = Keys.parse(sk.to_hex())
        signer = NostrSigner.keys(keys)
        database = NostrDatabase.sqlite("db/nostr_profiles.db")
        cli = ClientBuilder().signer(signer).database(database).opts(opts).build()

        cli.add_relay("wss://relay.damus.io")
        cli.add_relay("wss://nostr21.com")
        cli.connect()

        filter1 = Filter().kind(Kind(0))

        # filter = Filter().author(keys.public_key())
        print("Syncing Profile Database.. this might take a while..")
        dbopts = NegentropyOptions().direction(NegentropyDirection.DOWN)
        cli.reconcile(filter1, dbopts)
        print("Done Syncing Profile Database.")


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    dvm_config.SHOWLOG = True
    dvm_config.SCHEDULE_UPDATES_SECONDS = 600  # Every 10 seconds
    # Add NIP89
    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/981b560820bc283c58de7989b7abc6664996b487a531d852e4ef7322586a2122.jpg",
        "about": "I hunt down bot farms.",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "action": "mute",  # follow, unfollow, mute, unmute
        "nip90Params": {
            "max_results": {
                "required": False,
                "values": [],
                "description": "The number of maximum results to return (default currently 20)"
            }
        }
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    options = {"relay": "wss://relay.damus.io"}

    return DiscoveryBotFarms(name=name, dvm_config=dvm_config, nip89config=nip89config,
                      admin_config=admin_config, options=options)


if __name__ == '__main__':
    process_venv(DiscoveryBotFarms)



import json
import os
from datetime import timedelta
from nostr_sdk import Client, Timestamp, PublicKey, Tag, Keys, Options, SecretKey, NostrSigner, NostrDatabase, \
    ClientBuilder, Filter, NegentropyOptions, NegentropyDirection, init_logger, LogLevel, Event, EventId, Kind, \
    RelayOptions

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils import definitions
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config, check_and_set_d_tag_nip88, check_and_set_tiereventid_nip88
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag, create_amount_tag
from nostr_dvm.utils.output_utils import post_process_list_to_events

"""
This File contains a Module to discover popular notes
Accepted Inputs: none
Outputs: A list of events 
Params:  None
"""


class DicoverContentCurrentlyPopular(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_CONTENT_DISCOVERY
    TASK: str = "discover-content"
    FIX_COST: float = 0
    dvm_config: DVMConfig
    last_schedule: int
    db_since = 3600
    db_name = "db/nostr_recent_notes.db"
    min_reactions = 2
    personalized = False
    result = ""

    def __init__(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                 admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)
        super().__init__(name=name, dvm_config=dvm_config, nip89config=nip89config, nip88config=nip88config,
                         admin_config=admin_config, options=options)

        self.request_form = {"jobID": "generic"}
        opts = {
            "max_results": 200,
        }
        self.request_form['options'] = json.dumps(opts)

        self.last_schedule = Timestamp.now().as_secs()

        if self.options.get("db_name"):
            self.db_name = self.options.get("db_name")
        if self.options.get("db_since"):
            self.db_since = int(self.options.get("db_since"))


        use_logger = False
        if use_logger:
            init_logger(LogLevel.DEBUG)

        if self.dvm_config.UPDATE_DATABASE:
            self.sync_db()

        if not self.personalized:
            self.result = self.calculate_result(self.request_form)

    def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "text":
                    return False
        return True

    def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        self.dvm_config = dvm_config

        request_form = {"jobID": event.id().to_hex()}

        # default values
        search = ""
        max_results = 200

        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
            elif tag.as_vec()[0] == 'param':
                param = tag.as_vec()[1]
                if param == "max_results":  # check for param type
                    max_results = int(tag.as_vec()[2])

        options = {
            "max_results": max_results,
        }
        request_form['options'] = json.dumps(options)
        return request_form

    def process(self, request_form):
        # if the dvm supports individual results, recalculate it every time for the request
        if self.personalized:
            return self.calculate_result(request_form)
        # else return the result that gets updated once every scheduled update. In this case on database update.
        else:
            return self.result

    def calculate_result(self, request_form):
        from nostr_sdk import Filter
        from types import SimpleNamespace
        ns = SimpleNamespace()

        options = self.set_options(request_form)

        opts = (Options().wait_for_send(False).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT)))
        sk = SecretKey.from_hex(self.dvm_config.PRIVATE_KEY)
        keys = Keys.parse(sk.to_hex())
        signer = NostrSigner.keys(keys)

        database = NostrDatabase.sqlite(self.db_name)
        cli = ClientBuilder().database(database).signer(signer).opts(opts).build()
        cli.connect()

        # Negentropy reconciliation
        # Query events from database
        timestamp_hour_ago = Timestamp.now().as_secs() - self.db_since
        since = Timestamp.from_secs(timestamp_hour_ago)

        filter1 = Filter().kind(definitions.EventDefinitions.KIND_NOTE).since(since)
        events = cli.database().query([filter1])
        print("[" + self.dvm_config.NIP89.NAME + "] Considering " + str(len(events)) + " Events")
        ns.finallist = {}
        for event in events:
            if event.created_at().as_secs() > timestamp_hour_ago:
                filt = Filter().kinds([definitions.EventDefinitions.KIND_ZAP, definitions.EventDefinitions.KIND_REPOST,
                                       definitions.EventDefinitions.KIND_REACTION,
                                       definitions.EventDefinitions.KIND_NOTE]).event(event.id()).since(since)
                reactions = cli.database().query([filt])
                if len(reactions) >= self.min_reactions:
                    ns.finallist[event.id().to_hex()] = len(reactions)
        if len(ns.finallist) == 0:
            cli.disconnect()
            cli.shutdown()
            return self.result

        result_list = []
        finallist_sorted = sorted(ns.finallist.items(), key=lambda x: x[1], reverse=True)[:int(options["max_results"])]
        for entry in finallist_sorted:
            # print(EventId.parse(entry[0]).to_bech32() + "/" + EventId.parse(entry[0]).to_hex() + ": " + str(entry[1]))
            e_tag = Tag.parse(["e", entry[0]])
            result_list.append(e_tag.as_vec())
        cli.disconnect()
        cli.shutdown()
        return json.dumps(result_list)

    def post_process(self, result, event):
        """Overwrite the interface function to return a social client readable format, if requested"""
        for tag in event.tags():
            if tag.as_vec()[0] == 'output':
                format = tag.as_vec()[1]
                if format == "text/plain":  # check for output type
                    result = post_process_list_to_events(result)

        # if not text/plain, don't post-process
        return result

    def schedule(self, dvm_config):
        if dvm_config.SCHEDULE_UPDATES_SECONDS == 0:
            return 0
        else:
            if Timestamp.now().as_secs() >= self.last_schedule + dvm_config.SCHEDULE_UPDATES_SECONDS:
                if self.dvm_config.UPDATE_DATABASE:
                    self.sync_db()
                self.last_schedule = Timestamp.now().as_secs()
                self.result = self.calculate_result(self.request_form)
                return 1

    def sync_db(self):
        opts = (Options().wait_for_send(False).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT)))
        sk = SecretKey.from_hex(self.dvm_config.PRIVATE_KEY)
        keys = Keys.parse(sk.to_hex())
        signer = NostrSigner.keys(keys)
        database = NostrDatabase.sqlite(self.db_name)
        cli = ClientBuilder().signer(signer).database(database).opts(opts).build()

        ropts = RelayOptions().ping(True)
        cli.add_relay_with_opts("wss://relay.damus.io", ropts)
        cli.add_relay_with_opts("wss://nostr.oxtr.dev", ropts)
        cli.add_relay_with_opts("wss://nostr21.com", ropts)
        cli.connect()

        timestamp_hour_ago = Timestamp.now().as_secs() - self.db_since
        lasthour = Timestamp.from_secs(timestamp_hour_ago)

        filter1 = Filter().kinds([definitions.EventDefinitions.KIND_NOTE, definitions.EventDefinitions.KIND_REACTION,
                                  definitions.EventDefinitions.KIND_ZAP]).since(lasthour)  # Notes, reactions, zaps

        # filter = Filter().author(keys.public_key())
        print("[" + self.dvm_config.IDENTIFIER + "] Syncing notes of the last " + str(
            self.db_since) + " seconds.. this might take a while..")
        dbopts = NegentropyOptions().direction(NegentropyDirection.DOWN)
        cli.reconcile(filter1, dbopts)
        database.delete(Filter().until(Timestamp.from_secs(
            Timestamp.now().as_secs() - self.db_since)))  # Clear old events so db doesnt get too full.

        print(
            "[" + self.dvm_config.IDENTIFIER + "] Done Syncing Notes of the last " + str(self.db_since) + " seconds..")


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config, options,  cost=0, update_rate=180,  processing_msg=None, update_db=True):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    dvm_config.SHOWLOG = True
    dvm_config.SCHEDULE_UPDATES_SECONDS = update_rate  # Every 10 minutes
    dvm_config.UPDATE_DATABASE = update_db
    # Activate these to use a subscription based model instead
    # dvm_config.SUBSCRIPTION_REQUIRED = True
    # dvm_config.SUBSCRIPTION_DAILY_COST = 1
    dvm_config.FIX_COST = cost
    dvm_config.CUSTOM_PROCESSING_MESSAGE = processing_msg
    admin_config.LUD16 = dvm_config.LN_ADDRESS

    image = "https://image.nostr.build/b29b6ec4bf9b6184f69d33cb44862db0d90a2dd9a506532e7ba5698af7d36210.jpg",

    # Add NIP89
    nip89info = {
        "name": name,
        "image": image,
        "picture": image,
        "about": "I show notes that are currently popular",
        "lud16": dvm_config.LN_ADDRESS,
        "encryptionSupported": True,
        "cashuAccepted": True,
        "personalized": False,
        "amount": create_amount_tag(cost),
        "nip90Params": {
            "max_results": {
                "required": False,
                "values": [],
                "description": "The number of maximum results to return (default currently 100)"
            }
        }
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    #admin_config.UPDATE_PROFILE = False
    #admin_config.REBROADCAST_NIP89 = False

    return DicoverContentCurrentlyPopular(name=name, dvm_config=dvm_config, nip89config=nip89config,
                                          admin_config=admin_config, options=options)


def build_example_subscription(name, identifier, admin_config, options, update_rate=180, processing_msg=None, update_db=True):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    dvm_config.SHOWLOG = True
    dvm_config.SCHEDULE_UPDATES_SECONDS = update_rate  # Every 3 minutes
    dvm_config.UPDATE_DATABASE = update_db
    # Activate these to use a subscription based model instead
    # dvm_config.SUBSCRIPTION_DAILY_COST = 1
    dvm_config.FIX_COST = 0
    dvm_config.CUSTOM_PROCESSING_MESSAGE = processing_msg
    admin_config.LUD16 = dvm_config.LN_ADDRESS

    image = "https://image.nostr.build/b29b6ec4bf9b6184f69d33cb44862db0d90a2dd9a506532e7ba5698af7d36210.jpg",
    # Add NIP89
    nip89info = {
        "name": name,
        "image": image,
        "picture": image,
        "about": "I show notes that are currently popular all over Nostr. I'm also used for testing subscriptions.",
        "lud16": dvm_config.LN_ADDRESS,
        "encryptionSupported": True,
        "cashuAccepted": True,
        "subscription": True,
        "personalized": False,
        "nip90Params": {
            "max_results": {
                "required": False,
                "values": [],
                "description": "The number of maximum results to return (default currently 100)"
            }
        }
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    nip88config = NIP88Config()
    nip88config.DTAG = check_and_set_d_tag_nip88(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip88config.TIER_EVENT = check_and_set_tiereventid_nip88(identifier, "1")
    nip89config.NAME = name
    nip88config.IMAGE = nip89info["image"]
    nip88config.TITLE = name
    nip88config.AMOUNT_DAILY = 100
    nip88config.AMOUNT_MONTHLY = 2000
    nip88config.CONTENT = "Subscribe to the DVM for unlimited use during your subscription"
    nip88config.PERK1DESC = "Unlimited requests"
    nip88config.PERK2DESC = "Support NostrDVM & NostrSDK development"
    nip88config.PAYMENT_VERIFIER_PUBKEY = "5b5c045ecdf66fb540bdf2049fe0ef7f1a566fa427a4fe50d400a011b65a3a7e"

    #admin_config.UPDATE_PROFILE = False
    #admin_config.REBROADCAST_NIP89 = False
    #admin_config.REBROADCAST_NIP88 = False

    # admin_config.FETCH_NIP88 = True
    # admin_config.EVENTID = ""
    # admin_config.PRIVKEY = dvm_config.PRIVATE_KEY

    return DicoverContentCurrentlyPopular(name=name, dvm_config=dvm_config, nip89config=nip89config,
                                          nip88config=nip88config, options=options,
                                          admin_config=admin_config)


if __name__ == '__main__':
    process_venv(DicoverContentCurrentlyPopular)


import json
import os
from datetime import timedelta
from pathlib import Path
from typing import List

import dotenv
from nostr_sdk import Filter, Client, Alphabet, EventId, Event, PublicKey, Tag, Keys, nip04_decrypt, Metadata, Options, \
    Nip19Event, SingleLetterTag


def get_event_by_id(event_id: str, client: Client, config=None) -> Event | None:
    split = event_id.split(":")
    if len(split) == 3:
        pk = PublicKey.from_hex(split[1])
        id_filter = Filter().author(pk).custom_tag(SingleLetterTag.lowercase(Alphabet.D), [split[2]])
        events = client.get_events_of([id_filter], timedelta(seconds=config.RELAY_TIMEOUT))
    else:
        if str(event_id).startswith('note'):
            event_id = EventId.from_bech32(event_id)
        elif str(event_id).startswith("nevent"):
            event_id = Nip19Event.from_bech32(event_id).event_id()
        elif str(event_id).startswith('nostr:note'):
            event_id = EventId.from_nostr_uri(event_id)
        elif str(event_id).startswith("nostr:nevent"):
            event_id = Nip19Event.from_nostr_uri(event_id).event_id()

        else:
            event_id = EventId.from_hex(event_id)

        id_filter = Filter().id(event_id).limit(1)
        events = client.get_events_of([id_filter], timedelta(seconds=config.RELAY_TIMEOUT))
    if len(events) > 0:

        return events[0]
    else:
        return None


def get_events_by_ids(event_ids, client: Client, config=None) -> List | None:
    search_ids = []
    for event_id in event_ids:
        split = event_id.split(":")
        if len(split) == 3:
            pk = PublicKey.from_hex(split[1])
            id_filter = Filter().author(pk).custom_tag(SingleLetterTag.lowercase(Alphabet.D), [split[2]])
            events = client.get_events_of([id_filter], timedelta(seconds=config.RELAY_TIMEOUT))
        else:
            if str(event_id).startswith('note'):
                event_id = EventId.from_bech32(event_id)
            elif str(event_id).startswith("nevent"):
                event_id = Nip19Event.from_bech32(event_id).event_id()
            elif str(event_id).startswith('nostr:note'):
                event_id = EventId.from_nostr_uri(event_id)
            elif str(event_id).startswith("nostr:nevent"):
                event_id = Nip19Event.from_nostr_uri(event_id).event_id()

            else:
                event_id = EventId.from_hex(event_id)
        search_ids.append(event_id)

    id_filter = Filter().ids(search_ids)
    events = client.get_events_of([id_filter], timedelta(seconds=config.RELAY_TIMEOUT))
    if len(events) > 0:

        return events
    else:
        return None


def get_events_by_id(event_ids: list, client: Client, config=None) -> list[Event] | None:
    id_filter = Filter().ids(event_ids)
    events = client.get_events_of([id_filter], timedelta(seconds=config.RELAY_TIMEOUT))
    if len(events) > 0:
        return events
    else:
        return None


def get_referenced_event_by_id(event_id, client, dvm_config, kinds) -> Event | None:
    if kinds is None:
        kinds = []
    if str(event_id).startswith('note'):
        event_id = EventId.from_bech32(event_id)
    elif str(event_id).startswith("nevent"):
        event_id = Nip19Event.from_bech32(event_id).event_id()
    elif str(event_id).startswith('nostr:note'):
        event_id = EventId.from_nostr_uri(event_id)
    elif str(event_id).startswith("nostr:nevent"):
        event_id = Nip19Event.from_nostr_uri(event_id).event_id()
    else:
        event_id = EventId.from_hex(event_id)

    if len(kinds) > 0:
        job_id_filter = Filter().kinds(kinds).event(event_id).limit(1)
    else:
        job_id_filter = Filter().event(event_id).limit(1)

    events = client.get_events_of([job_id_filter], timedelta(seconds=dvm_config.RELAY_TIMEOUT))

    if len(events) > 0:
        return events[0]
    else:
        return None


def send_event(event: Event, client: Client, dvm_config) -> EventId:
    try:
        relays = []

        for tag in event.tags():
            if tag.as_vec()[0] == 'relays':
                for index, param in enumerate(tag.as_vec()):
                    if index != 0:
                        relays.append(tag.as_vec()[index])

        for relay in relays:
            if relay not in dvm_config.RELAY_LIST:
                client.add_relay(relay)

        event_id = client.send_event(event)

        for relay in relays:
            if relay not in dvm_config.RELAY_LIST:
                client.remove_relay(relay)

        return event_id
    except Exception as e:
        print(e)


def check_and_decrypt_tags(event, dvm_config):
    try:

        is_encrypted = False
        p = ""
        for tag in event.tags():
            if tag.as_vec()[0] == 'encrypted':
                is_encrypted = True
            elif tag.as_vec()[0] == 'p':
                p = tag.as_vec()[1]

        if is_encrypted:
            if p != dvm_config.PUBLIC_KEY:
                print("[" + dvm_config.NIP89.NAME + "] Task encrypted and not addressed to this DVM, "
                                                    "skipping..")
                return None

            elif p == dvm_config.PUBLIC_KEY:
                tags_str = nip04_decrypt(Keys.parse(dvm_config.PRIVATE_KEY).secret_key(),
                                         event.author(), event.content())
                params = json.loads(tags_str)
                params.append(Tag.parse(["p", p]).as_vec())
                params.append(Tag.parse(["encrypted"]).as_vec())
                event_as_json = json.loads(event.as_json())
                event_as_json['tags'] = params
                event_as_json['content'] = ""
                event = Event.from_json(json.dumps(event_as_json))
    except Exception as e:
        print(e)

    return event


def check_and_decrypt_own_tags(event, dvm_config):
    try:
        is_encrypted = False
        p = ""
        for tag in event.tags():
            if tag.as_vec()[0] == 'encrypted':
                is_encrypted = True
            elif tag.as_vec()[0] == 'p':
                p = tag.as_vec()[1]

        if is_encrypted:
            if dvm_config.PUBLIC_KEY != event.author().to_hex():
                print("[" + dvm_config.NIP89.NAME + "] Task encrypted and not addressed to this DVM, "
                                                    "skipping..")
                return None

            elif event.author().to_hex() == dvm_config.PUBLIC_KEY:
                tags_str = nip04_decrypt(Keys.parse(dvm_config.PRIVATE_KEY).secret_key(),
                                         PublicKey.from_hex(p), event.content())
                params = json.loads(tags_str)
                params.append(Tag.parse(["p", p]).as_vec())
                params.append(Tag.parse(["encrypted"]).as_vec())
                event_as_json = json.loads(event.as_json())
                event_as_json['tags'] = params
                event_as_json['content'] = ""
                event = Event.from_json(json.dumps(event_as_json))
    except Exception as e:
        print(e)

    return event


def update_profile(dvm_config, client, lud16=""):
    keys = Keys.parse(dvm_config.PRIVATE_KEY)
    nip89content = json.loads(dvm_config.NIP89.CONTENT)
    if nip89content.get("name"):
        name = nip89content.get("name")
        about = nip89content.get("about")
        image = nip89content.get("image")

        # Set metadata
        metadata = Metadata() \
            .set_name(name) \
            .set_display_name(name) \
            .set_about(about) \
            .set_picture(image) \
            .set_lud16(lud16) \
            .set_nip05(lud16)
        # .set_banner("https://example.com/banner.png") \
        print("[" + dvm_config.NIP89.NAME + "] Setting profile metadata for " + keys.public_key().to_bech32() + "...")
        print(metadata.as_json())
        client.set_metadata(metadata)


def check_and_set_private_key(identifier):
    if not os.getenv("DVM_PRIVATE_KEY_" + identifier.upper()):
        pk = Keys.generate().secret_key().to_hex()
        add_pk_to_env_file("DVM_PRIVATE_KEY_" + identifier.upper(), pk)
        return pk
    else:
        return os.getenv("DVM_PRIVATE_KEY_" + identifier.upper())


def add_pk_to_env_file(dtag, oskey):
    env_path = Path('.env')
    if env_path.is_file():
        print(f'loading environment from {env_path.resolve()}')
        dotenv.load_dotenv(env_path, verbose=True, override=True)
        dotenv.set_key(env_path, dtag, oskey)



import json
import os
from datetime import timedelta
from nostr_sdk import Client, Timestamp, PublicKey, Tag, Keys, Options, SecretKey, NostrSigner, Kind, RelayOptions

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag
from nostr_dvm.utils.output_utils import post_process_list_to_events

"""
This File contains a Module to search for notes
Accepted Inputs: a search query
Outputs: A list of events 
Params:  None
"""


class AdvancedSearch(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_CONTENT_SEARCH
    TASK: str = "search-content"
    FIX_COST: float = 0
    dvm_config: DVMConfig

    def __init__(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                 admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)
        super().__init__(name=name, dvm_config=dvm_config, nip89config=nip89config, nip88config=nip88config,
                         admin_config=admin_config, options=options)

    def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "text":
                    return False
        return True

    def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        self.dvm_config = dvm_config
        print(self.dvm_config.PRIVATE_KEY)

        request_form = {"jobID": event.id().to_hex()}

        # default values
        user = ""
        users = []
        since_seconds = Timestamp.now().as_secs() - (365 * 24 * 60 * 60)
        until_seconds = Timestamp.now().as_secs()
        search = ""
        max_results = 100
        relay = "wss://relay.nostr.band"
        
        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                if input_type == "text":
                    search = tag.as_vec()[1]
            elif tag.as_vec()[0] == 'param':
                param = tag.as_vec()[1]
                if param == "user":  # check for param type
                    #user = tag.as_vec()[2]
                    users.append(Tag.parse(["p", tag.as_vec()[2]]))
                elif param == "users":  # check for param type
                    users = json.loads(tag.as_vec()[2])
                elif param == "since":  # check for param type
                    since_seconds = int(tag.as_vec()[2])
                elif param == "until":  # check for param type
                    until_seconds = min(int(tag.as_vec()[2]), until_seconds)
                elif param == "max_results":  # check for param type
                    max_results = int(tag.as_vec()[2])

        options = {
            "search": search,
            "users": users,
            "since": since_seconds,
            "until": until_seconds,
            "max_results": max_results,
            "relay": relay
        }
        request_form['options'] = json.dumps(options)
        return request_form

    def process(self, request_form):
        from nostr_sdk import Filter
        options = self.set_options(request_form)

        opts = (Options().wait_for_send(False).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT)))
        sk = SecretKey.from_hex(self.dvm_config.PRIVATE_KEY)
        keys = Keys.parse(sk.to_hex())
        signer = NostrSigner.keys(keys)
        cli = Client.with_opts(signer, opts)

        ropts = RelayOptions().ping(False)
        cli.add_relay_with_opts(options["relay"], ropts)

        cli.connect()

        #earch_since_seconds = int(options["since"]) * 24 * 60 * 60
        #dif = Timestamp.now().as_secs() - search_since_seconds
        #search_since = Timestamp.from_secs(dif)
        search_since = Timestamp.from_secs(int(options["since"]))

        #search_until_seconds = int(options["until"]) * 24 * 60 * 60
        #dif = Timestamp.now().as_secs() - search_until_seconds
        #search_until = Timestamp.from_secs(dif)
        search_until = Timestamp.from_secs(int(options["until"]))
        userkeys = []
        for user in options["users"]:
            tag = Tag.parse(user)
            user = tag.as_vec()[1]
            #user = user[1]
            user = str(user).lstrip("@")
            if str(user).startswith('npub'):
                userkey = PublicKey.from_bech32(user)
            elif str(user).startswith("nostr:npub"):
                userkey = PublicKey.from_nostr_uri(user)
            else:
                userkey = PublicKey.from_hex(user)

            userkeys.append(userkey)

        if not options["users"]:
            notes_filter = Filter().kind(Kind(1)).search(options["search"]).since(search_since).until(search_until).limit(options["max_results"])
        elif options["search"] == "":
                notes_filter = Filter().kind(Kind(1)).authors(userkeys).since(search_since).until(
                    search_until).limit(options["max_results"])
        else:
                notes_filter = Filter().kind(Kind(1)).authors(userkeys).search(options["search"]).since(
                    search_since).until(search_until).limit(options["max_results"])


        events = cli.get_events_of([notes_filter], timedelta(seconds=5))

        result_list = []
        if len(events) > 0:

            for event in events:
                e_tag = Tag.parse(["e", event.id().to_hex()])
                #print(e_tag.as_vec())
                result_list.append(e_tag.as_vec())

        return json.dumps(result_list)

    def post_process(self, result, event):
        """Overwrite the interface function to return a social client readable format, if requested"""
        for tag in event.tags():
            if tag.as_vec()[0] == 'output':
                format = tag.as_vec()[1]
                if format == "text/plain":  # check for output type
                    result = post_process_list_to_events(result)

        # if not text/plain, don't post-process
        return result


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    # Add NIP89
    nip89info = {
        "name": name,
        "image": "https://nostr.band/android-chrome-192x192.png",
        "about": "I search notes on Nostr.band.",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {
            "users": {
                "required": False,
                "values": [],
                "description": "Search for content from specific users"
            },
            "since": {
                "required": False,
                "values": [],
                "description": "A unix timestamp in the past from where the search should start"
            },
            "until": {
                "required": False,
                "values": [],
                "description": "A unix timestamp that tells until the search should include results"
            },
            "max_results": {
                "required": False,
                "values": [],
                "description": "The number of maximum results to return (default currently 20)"
            }
        }
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    options = {"relay": "wss://relay.nostr.band"}

    return AdvancedSearch(name=name, dvm_config=dvm_config, nip89config=nip89config,
                          admin_config=admin_config, options=options)


if __name__ == '__main__':
    process_venv(AdvancedSearch)



import json
import os
import signal
import time
from datetime import timedelta

from nostr_sdk import (Keys, Client, Timestamp, Filter, nip04_decrypt, HandleNotification, EventBuilder, PublicKey,
                       Options, Tag, Event, nip04_encrypt, NostrSigner, EventId, Nip19Event, Kind, KindEnum,
                       UnsignedEvent, UnwrappedGift)

from nostr_dvm.utils.admin_utils import admin_make_database_updates
from nostr_dvm.utils.database_utils import get_or_add_user, update_user_balance, create_sql_table, update_sql_table
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.nip89_utils import nip89_fetch_events_pubkey, NIP89Config
from nostr_dvm.utils.nostr_utils import send_event
from nostr_dvm.utils.output_utils import PostProcessFunctionType, post_process_list_to_users, \
    post_process_list_to_events
from nostr_dvm.utils.zap_utils import parse_zap_event_tags, pay_bolt11_ln_bits, zaprequest
from nostr_dvm.utils.cashu_utils import redeem_cashu


class Bot:
    job_list: list

    # This is a simple list just to keep track which events we created and manage, so we don't pay for other requests
    def __init__(self, dvm_config, admin_config=None):
        self.NAME = "Bot"
        dvm_config.DB = "db/" + self.NAME + ".db"
        self.dvm_config = dvm_config
        nip89config = NIP89Config()
        nip89config.NAME = self.NAME
        self.dvm_config.NIP89 = nip89config
        self.admin_config = admin_config
        self.keys = Keys.parse(dvm_config.PRIVATE_KEY)
        wait_for_send = True
        skip_disconnected_relays = True
        opts = (Options().wait_for_send(wait_for_send).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT))
                .skip_disconnected_relays(skip_disconnected_relays))
        signer = NostrSigner.keys(self.keys)
        self.client = Client.with_opts(signer, opts)

        pk = self.keys.public_key()

        self.job_list = []

        print("Nostr BOT public key: " + str(pk.to_bech32()) + " Hex: " + str(pk.to_hex()) + " Name: " + self.NAME +
              " Supported DVM tasks: " +
              ', '.join(p.NAME + ":" + p.TASK for p in self.dvm_config.SUPPORTED_DVMS) + "\n")

        for relay in self.dvm_config.RELAY_LIST:
            self.client.add_relay(relay)
        self.client.connect()

        zap_filter = Filter().pubkey(pk).kinds([EventDefinitions.KIND_ZAP]).since(Timestamp.now())
        dm_filter = Filter().pubkey(pk).kinds([EventDefinitions.KIND_DM]).since(Timestamp.now())
        nip59_filter = Filter().pubkey(pk).kind(Kind.from_enum(KindEnum.GIFT_WRAP())).since(
            Timestamp.from_secs(Timestamp.now().as_secs() - 60 * 60 * 24 * 7))
        kinds = [EventDefinitions.KIND_NIP90_GENERIC, EventDefinitions.KIND_FEEDBACK]
        for dvm in self.dvm_config.SUPPORTED_DVMS:
            if dvm.KIND not in kinds:
                kinds.append(Kind(dvm.KIND.as_u64() + 1000))
        dvm_filter = (Filter().kinds(kinds).since(Timestamp.now()))

        self.client.subscribe([zap_filter, dm_filter, nip59_filter, dvm_filter], None)

        create_sql_table(self.dvm_config.DB)
        admin_make_database_updates(adminconfig=self.admin_config, dvmconfig=self.dvm_config, client=self.client)

        # add_sql_table_column(dvm_config.DB)

        class NotificationHandler(HandleNotification):
            client = self.client
            dvm_config = self.dvm_config
            keys = self.keys

            def handle(self, relay_url, subscription_id, nostr_event):
                if (EventDefinitions.KIND_NIP90_EXTRACT_TEXT.as_u64() + 1000 <= nostr_event.kind().as_u64()
                        <= EventDefinitions.KIND_NIP90_GENERIC.as_u64() + 1000):
                    handle_nip90_response_event(nostr_event)
                elif nostr_event.kind() == EventDefinitions.KIND_FEEDBACK:
                    handle_nip90_feedback(nostr_event)

                elif nostr_event.kind() == EventDefinitions.KIND_ZAP:
                    handle_zap(nostr_event)

                elif nostr_event.kind() == EventDefinitions.KIND_DM:
                    try:
                        handle_dm(nostr_event, False)
                    except Exception as e:
                        print(f"Error during content NIP04 decryption: {e}")
                elif nostr_event.kind().match_enum(KindEnum.GIFT_WRAP()):
                    try:
                      handle_dm(nostr_event, True)
                    except Exception as e:
                        print(f"Error during content NIP59 decryption: {e}")

            def handle_msg(self, relay_url, msg):
                return

        def handle_dm(nostr_event, giftwrap):
            sender = nostr_event.author().to_hex()
            if sender == self.keys.public_key().to_hex():
                return
            decrypted_text = ""
            try:
                sealed = " "
                if giftwrap:
                    try:
                        # Extract rumor
                        unwrapped_gift = UnwrappedGift.from_gift_wrap(self.keys, nostr_event)
                        sender = unwrapped_gift.sender().to_hex()
                        rumor: UnsignedEvent = unwrapped_gift.rumor()

                        # Check timestamp of rumor
                        if rumor.created_at().as_secs() >= Timestamp.now().as_secs():
                            if rumor.kind().match_enum(KindEnum.SEALED_DIRECT()):
                                decrypted_text = rumor.content()
                                print(f"Received new msg [sealed]: {decrypted_text}")
                                sealed = " [sealed] "
                                # client.send_sealed_msg(sender, f"Echo: {msg}", None)
                            else:
                                print(f"{rumor.as_json()}")
                    except Exception as e:
                        print(f"Error during content NIP59 decryption: {e}")

                else:
                    try:
                        decrypted_text = nip04_decrypt(self.keys.secret_key(), nostr_event.author(),
                                                       nostr_event.content())
                    except Exception as e:
                        print(f"Error during content NIP04 decryption: {e}")

                if decrypted_text != "":
                    user = get_or_add_user(db=self.dvm_config.DB, npub=sender, client=self.client,
                                           config=self.dvm_config)

                    print("[" + self.NAME + "]" + sealed + "Message from " + user.name + ": " + decrypted_text)

                    # if user selects an index from the overview list...
                    if decrypted_text != "" and decrypted_text[0].isdigit():

                        split = decrypted_text.split(' ')
                        index = int(split[0]) - 1
                        # if user sends index info, e.g. 1 info, we fetch the nip89 information and reply with it.
                        if len(split) > 1 and split[1].lower() == "info":
                            answer_nip89(nostr_event, index, giftwrap, sender)
                        # otherwise we probably have to do some work, so build an event from input and send it to the DVM
                        else:
                            task = self.dvm_config.SUPPORTED_DVMS[index].TASK
                            print("[" + self.NAME + "] Request from " + str(user.name) + " (" + str(user.nip05) +
                                  ", Balance: " + str(user.balance) + " Sats) Task: " + str(task))

                            if user.isblacklisted:
                                # If users are blacklisted for some reason, tell them.
                                answer_blacklisted(nostr_event, giftwrap, sender)

                            else:
                                # Parse inputs to params
                                tags = build_params(decrypted_text, sender, index)
                                p_tag = Tag.parse(['p', self.dvm_config.SUPPORTED_DVMS[index].PUBLIC_KEY])

                                if self.dvm_config.SUPPORTED_DVMS[index].SUPPORTS_ENCRYPTION:
                                    tags_str = []
                                    for tag in tags:
                                        tags_str.append(tag.as_vec())
                                    params_as_str = json.dumps(tags_str)
                                    print(params_as_str)
                                    #  and encrypt them
                                    encrypted_params = nip04_encrypt(self.keys.secret_key(),
                                                                     PublicKey.from_hex(
                                                                         self.dvm_config.SUPPORTED_DVMS[
                                                                             index].PUBLIC_KEY),
                                                                     params_as_str)
                                    #  add encrypted and p tag on the outside
                                    encrypted_tag = Tag.parse(['encrypted'])
                                    #  add the encrypted params to the content
                                    nip90request = (EventBuilder(self.dvm_config.SUPPORTED_DVMS[index].KIND,
                                                                 encrypted_params, [p_tag, encrypted_tag]).
                                                    to_event(self.keys))
                                else:
                                    tags.append(p_tag)

                                    nip90request = (EventBuilder(self.dvm_config.SUPPORTED_DVMS[index].KIND,
                                                                 "", tags).
                                                    to_event(self.keys))

                                # remember in the job_list that we have made an event, if anybody asks for payment,
                                # we know we actually sent the request
                                entry = {"npub": user.npub, "event_id": nip90request.id().to_hex(),
                                         "dvm_key": self.dvm_config.SUPPORTED_DVMS[index].PUBLIC_KEY, "is_paid": False, "giftwrap": giftwrap}
                                self.job_list.append(entry)

                                # send the event to the DVM
                                send_event(nip90request, client=self.client, dvm_config=self.dvm_config)
                                # print(nip90request.as_json())



                    elif decrypted_text.lower().startswith("balance"):
                        time.sleep(3.0)
                        message = "Your current balance is " + str(user.balance) + ("Sats. Zap me to add to your "
                                                                                    "balance. I will use your "
                                                                                    "balance interact with the DVMs "
                                                                                    "for you.\n I support both "
                                                                                    "public and private Zaps, "
                                                                                    "as well as "
                                                                                    "Zapplepay.\nAlternativly you "
                                                                                    "can add a #cashu token with "
                                                                                    "\"-cashu cashuASomeToken\" to "
                                                                                    "your command.\n Make sure the "
                                                                                    "token is worth the requested "
                                                                                    "amount mint fees (at least 3 "
                                                                                    "sat).\n Not all DVMs might "
                                                                                    "accept Cashu tokens.")
                        if giftwrap:
                            self.client.send_sealed_msg(PublicKey.parse(sender), message, None)
                        else:
                            evt = EventBuilder.encrypted_direct_msg(self.keys, PublicKey.parse(sender),
                                                                    message,None).to_event(self.keys)
                            send_event(evt, client=self.client, dvm_config=dvm_config)
                    elif decrypted_text.startswith("cashuA"):
                        print("Received Cashu token:" + decrypted_text)
                        cashu_redeemed, cashu_message, total_amount, fees = redeem_cashu(decrypted_text,
                                                                                         self.dvm_config,
                                                                                         self.client)
                        print(cashu_message)
                        if cashu_message == "success":
                            update_user_balance(self.dvm_config.DB, sender, total_amount, client=self.client,
                                                config=self.dvm_config)
                        else:
                            time.sleep(2.0)
                            message = "Error: " + cashu_message + ". Token has not been redeemed."

                            if giftwrap:
                                self.client.send_sealed_msg(PublicKey.parse(sender), message, None)
                            else:
                                evt = EventBuilder.encrypted_direct_msg(self.keys, PublicKey.from_hex(sender), message,
                                                                        None).to_event(self.keys)
                                send_event(evt, client=self.client, dvm_config=self.dvm_config)
                    elif decrypted_text.lower().startswith("what's the second best"):
                        time.sleep(3.0)
                        message = "No, there is no second best.\n\nhttps://cdn.nostr.build/p/mYLv.mp4"
                        if giftwrap:
                            self.client.send_sealed_msg(PublicKey.parse(sender), message, None)
                        else:
                            evt = EventBuilder.encrypted_direct_msg(self.keys, PublicKey.parse(sender),
                                                                    message,
                                                                    nostr_event.id()).to_event(self.keys)
                            send_event(evt, client=self.client, dvm_config=self.dvm_config)

                    else:
                        # Build an overview of known DVMs and send it to the user
                        answer_overview(nostr_event, giftwrap, sender)

            except Exception as e:
                print("Error in bot " + str(e))

        def handle_nip90_feedback(nostr_event):
            # print(nostr_event.as_json())
            try:
                is_encrypted = False
                status = ""
                etag = ""
                ptag = ""
                content = nostr_event.content()
                for tag in nostr_event.tags():
                    if tag.as_vec()[0] == "status":
                        status = tag.as_vec()[1]
                        if len(tag.as_vec()) > 2:
                            content = tag.as_vec()[2]
                    elif tag.as_vec()[0] == "e":
                        etag = tag.as_vec()[1]
                    elif tag.as_vec()[0] == "p":
                        ptag = tag.as_vec()[1]
                    elif tag.as_vec()[0] == "encrypted":
                        is_encrypted = True

                if is_encrypted:
                    if ptag == self.keys.public_key().to_hex():
                        tags_str = nip04_decrypt(Keys.parse(dvm_config.PRIVATE_KEY).secret_key(),
                                                 nostr_event.author(), nostr_event.content())
                        params = json.loads(tags_str)
                        params.append(Tag.parse(["p", ptag]).as_vec())
                        params.append(Tag.parse(["encrypted"]).as_vec())
                        event_as_json = json.loads(nostr_event.as_json())
                        event_as_json['tags'] = params
                        event_as_json['content'] = ""
                        nostr_event = Event.from_json(json.dumps(event_as_json))

                        for tag in nostr_event.tags():
                            if tag.as_vec()[0] == "status":
                                status = tag.as_vec()[1]
                                if len(tag.as_vec()) > 2:
                                    content = tag.as_vec()[2]
                            elif tag.as_vec()[0] == "e":
                                etag = tag.as_vec()[1]
                            elif tag.as_vec()[0] == "content":
                                content = tag.as_vec()[1]

                    else:
                        return

                if status == "success" or status == "error" or status == "processing" or status == "partial" and content != "":
                    entry = next((x for x in self.job_list if x['event_id'] == etag), None)
                    if entry is not None and entry['dvm_key'] == nostr_event.author().to_hex():
                        user = get_or_add_user(db=self.dvm_config.DB, npub=entry['npub'],
                                               client=self.client, config=self.dvm_config)
                        time.sleep(2.0)
                        if entry["giftwrap"]:
                            self.client.send_sealed_msg(PublicKey.parse(entry["npub"]), content, None)
                        else:
                            reply_event = EventBuilder.encrypted_direct_msg(self.keys,
                                                                            PublicKey.from_hex(entry['npub']),
                                                                            content,
                                                                            None).to_event(self.keys)

                            send_event(reply_event, client=self.client, dvm_config=dvm_config)
                        print(status + ": " + content)
                        print(
                            "[" + self.NAME + "] Received reaction from " + nostr_event.author().to_hex() + " message to orignal sender " + user.name)

                elif status == "payment-required" or status == "partial":
                    for tag in nostr_event.tags():
                        if tag.as_vec()[0] == "amount":
                            amount_msats = int(tag.as_vec()[1])
                            amount = int(amount_msats / 1000)
                            entry = next((x for x in self.job_list if x['event_id'] == etag), None)
                            if entry is not None and entry['is_paid'] is False and entry[
                                'dvm_key'] == nostr_event.author().to_hex():
                                # if we get a bolt11, we pay and move on
                                user = get_or_add_user(db=self.dvm_config.DB, npub=entry["npub"],
                                                       client=self.client, config=self.dvm_config)
                                if user.balance >= amount:
                                    balance = max(user.balance - amount, 0)
                                    update_sql_table(db=self.dvm_config.DB, npub=user.npub, balance=balance,
                                                     iswhitelisted=user.iswhitelisted, isblacklisted=user.isblacklisted,
                                                     nip05=user.nip05, lud16=user.lud16, name=user.name,
                                                     lastactive=Timestamp.now().as_secs(), subscribed=user.subscribed)

                                    message = "Paid " + str(amount) + " Sats from balance to DVM. New balance is " + str(balance) + " Sats.\n"
                                    if entry["giftwrap"]:
                                        self.client.send_sealed_msg(PublicKey.parse(entry["npub"]), message, None)
                                    else:
                                        evt = EventBuilder.encrypted_direct_msg(self.keys,
                                                                                PublicKey.parse(entry["npub"]),
                                                                                message,
                                                                                None).to_event(self.keys)
                                        send_event(evt, client=self.client, dvm_config=dvm_config)
                                    print(
                                        "[" + self.NAME + "] Replying " + user.name + " with \"scheduled\" confirmation")

                                else:
                                    print("Bot payment-required")
                                    time.sleep(2.0)
                                    evt = EventBuilder.encrypted_direct_msg(self.keys,
                                                                            PublicKey.parse(entry["npub"]),
                                                                            "Current balance: " + str(
                                                                                user.balance) + " Sats. Balance of " + str(
                                                                                amount) + " Sats required. Please zap me with at least " +
                                                                            str(int(amount - user.balance))
                                                                            + " Sats, then try again.",
                                                                            None).to_event(self.keys)
                                    send_event(evt, client=self.client, dvm_config=dvm_config)
                                    return

                                if len(tag.as_vec()) > 2:
                                    bolt11 = tag.as_vec()[2]
                                # else we create a zap
                                else:
                                    user = get_or_add_user(db=self.dvm_config.DB, npub=nostr_event.author().to_hex(),
                                                           client=self.client, config=self.dvm_config)
                                    print("Paying: " + user.name)
                                    bolt11 = zaprequest(user.lud16, amount, "Zap", nostr_event, self.keys,
                                                        self.dvm_config,
                                                        "private")
                                    if bolt11 == None:
                                        print("Receiver has no Lightning address")
                                        return
                                try:
                                    print(bolt11)
                                    payment_hash = pay_bolt11_ln_bits(bolt11, self.dvm_config)
                                    self.job_list[self.job_list.index(entry)]['is_paid'] = True
                                    print("[" + self.NAME + "] payment_hash: " + payment_hash +
                                          " Forwarding payment of " + str(amount) + " Sats to DVM")
                                except Exception as e:
                                    print(e)


            except Exception as e:
                print(e)

        def handle_nip90_response_event(nostr_event: Event):
            try:
                ptag = ""
                etag = ""
                is_encrypted = False
                for tag in nostr_event.tags():
                    if tag.as_vec()[0] == "e":
                        etag = tag.as_vec()[1]
                    elif tag.as_vec()[0] == "p":
                        ptag = tag.as_vec()[1]
                    elif tag.as_vec()[0] == "encrypted":
                        is_encrypted = True

                entry = next((x for x in self.job_list if x['event_id'] == etag), None)
                if entry is not None and entry[
                    'dvm_key'] == nostr_event.author().to_hex():
                    print(entry)
                    user = get_or_add_user(db=self.dvm_config.DB, npub=entry['npub'],
                                           client=self.client, config=self.dvm_config)

                    self.job_list.remove(entry)
                    content = nostr_event.content()
                    if is_encrypted:
                        if ptag == self.keys.public_key().to_hex():
                            content = nip04_decrypt(self.keys.secret_key(), nostr_event.author(), content)
                        else:
                            return

                    dvms = [x for x in self.dvm_config.SUPPORTED_DVMS if
                            x.PUBLIC_KEY == nostr_event.author().to_hex() and x.KIND.as_u64() == nostr_event.kind().as_u64() - 1000]
                    if len(dvms) > 0:
                        dvm = dvms[0]
                        if dvm.dvm_config.EXTERNAL_POST_PROCESS_TYPE != PostProcessFunctionType.NONE:
                            if dvm.dvm_config.EXTERNAL_POST_PROCESS_TYPE == PostProcessFunctionType.LIST_TO_EVENTS:
                                content = post_process_list_to_events(content)
                            elif dvm.dvm_config.EXTERNAL_POST_PROCESS_TYPE == PostProcessFunctionType.LIST_TO_USERS:
                                content = post_process_list_to_users(content)

                    print("[" + self.NAME + "] Received results, message to orignal sender " + user.name)
                    time.sleep(1.0)
                    if entry["giftwrap"]:
                        self.client.send_sealed_msg(PublicKey.parse(user.npub), content, None)
                    else:
                        reply_event = EventBuilder.encrypted_direct_msg(self.keys,
                                                                        PublicKey.parse(user.npub),
                                                                        content,
                                                                        None).to_event(self.keys)
                        send_event(reply_event, client=self.client, dvm_config=dvm_config)

            except Exception as e:
                print(e)

        def handle_zap(zap_event):
            print("[" + self.NAME + "] Zap received")
            try:
                invoice_amount, zapped_event, sender, message, anon = parse_zap_event_tags(zap_event,
                                                                                           self.keys, self.NAME,
                                                                                           self.client, self.dvm_config)

                etag = ""
                print(zap_event.tags())
                print(zapped_event.tags())
                for tag in zapped_event.tags():
                    if tag.as_vec()[0] == "e":
                        etag = tag.as_vec()[1]

                user = get_or_add_user(self.dvm_config.DB, sender, client=self.client, config=self.dvm_config)

                entry = next((x for x in self.job_list if x['event_id'] == etag), None)
                print(entry)
                # print(entry['dvm_key'])
                # print(str(zapped_event.author().to_hex()))
                # print(str(zap_event.author().to_hex()))
                print(sender)
                if entry is not None and entry['is_paid'] is True and entry['dvm_key'] == sender:
                    # if we get a bolt11, we pay and move on
                    user = get_or_add_user(db=self.dvm_config.DB, npub=entry["npub"],
                                           client=self.client, config=self.dvm_config)

                    sender = user.npub

                if zapped_event is not None:
                    if not anon:
                        print("[" + self.NAME + "] Note Zap received for Bot balance: " + str(
                            invoice_amount) + " Sats from " + str(
                            user.name))
                        update_user_balance(self.dvm_config.DB, sender, invoice_amount, client=self.client,
                                            config=self.dvm_config)

                        # a regular note
                elif not anon:
                    print("[" + self.NAME + "] Profile Zap received for Bot balance: " + str(
                        invoice_amount) + " Sats from " + str(
                        user.name))
                    update_user_balance(self.dvm_config.DB, sender, invoice_amount, client=self.client,
                                        config=self.dvm_config)

            except Exception as e:
                print("[" + self.NAME + "] Error during content decryption:" + str(e))

        def answer_overview(nostr_event, giftwrap, sender):
            message = "DVMs that I support:\n\n"
            index = 1
            for p in self.dvm_config.SUPPORTED_DVMS:
                if p.PER_UNIT_COST != 0 and p.PER_UNIT_COST is not None:
                    message += (str(index) + " " + p.NAME + " " + p.TASK + "\n\t" + str(p.FIX_COST) +
                                " Sats + " + str(p.PER_UNIT_COST) + " Sats per Second\n\n")
                else:
                    message += (str(index) + " " + p.NAME + " " + p.TASK + "\n\t" + str(p.FIX_COST) +
                                " Sats\n\n")
                index += 1

            time.sleep(3.0)

            text =  message + "\nSelect an Index and provide an input (e.g. \"2 A purple ostrich\")\nType \"index info\" to learn more about each DVM. (e.g. \"2 info\")\n\n Type \"balance\" to see your current balance"
            if giftwrap:
                self.client.send_sealed_msg(PublicKey.parse(sender), text, None)
            else:
                evt = EventBuilder.encrypted_direct_msg(self.keys, PublicKey.parse(sender),
                                                        text,
                                                        nostr_event.id()).to_event(self.keys)

                send_event(evt, client=self.client, dvm_config=dvm_config)

        def answer_blacklisted(nostr_event, giftwrap, sender):
            message = "Your are currently blocked from this service."
            if giftwrap:
                self.client.send_sealed_msg(PublicKey.parse(sender), message, None)
            else:
                 # For some reason an admin might blacklist npubs, e.g. for abusing the service
                evt = EventBuilder.encrypted_direct_msg(self.keys, nostr_event.author(),
                                                        message, None).to_event(self.keys)
                send_event(evt, client=self.client, dvm_config=dvm_config)

        def answer_nip89(nostr_event, index, giftwrap, sender):
            info = print_dvm_info(self.client, index)
            if info is None:
                info =  "No NIP89 Info found for " + self.dvm_config.SUPPORTED_DVMS[index].NAME
            time.sleep(2.0)

            if giftwrap:
                self.client.send_sealed_msg(PublicKey.parse(sender), info, None)
            else:
                    evt = EventBuilder.encrypted_direct_msg(self.keys, nostr_event.author(),
                                                            info, None).to_event(self.keys)
                    send_event(evt, client=self.client, dvm_config=dvm_config)

        def build_params(decrypted_text, author, index):
            tags = []
            splitzero = decrypted_text.split(' -')
            split = splitzero[0].split(' ')
            # If only a command without parameters is sent, we assume no input is required, and that means the dvm might take in the user as input (e.g. for content discovery)
            if len(split) == 1:
                remaining_text = decrypted_text.replace(split[0], "")
                params = remaining_text.split(" -")
                tag = Tag.parse(["param", "user", author])
                tags.append(tag)
                for i in params:
                    print(i)
                    if i != " ":
                        try:
                            split = i.split(" ")
                            if len(split) > 1:
                                param = str(split[0])
                                print(str(param))
                                value = str(split[1])
                                print(str(value))
                                if param == "cashu":
                                    tag = Tag.parse([param, value])
                                else:
                                    if param == "user":
                                        if value.startswith("@") or value.startswith("nostr:") or value.startswith(
                                                "npub"):
                                            value = PublicKey.from_bech32(
                                                value.replace("@", "").replace("nostr:", "")).to_hex()
                                    tag = Tag.parse(["param", param, value])
                                tags.append(tag)
                        except Exception as e:
                            print(e)
                            print("Couldn't add " + str(i))
                        output = Tag.parse(["output", "text/plain"])
                        tags.append(output)
                        relay_list = ["relays"]
                        for relay in self.dvm_config.RELAY_LIST:
                            relay_list.append(relay)
                            relays = Tag.parse(relay_list)
                            tags.append(relays)

                return tags

            tags = []
            command = decrypted_text.replace(split[0] + " ", "")
            split = command.split(" -")
            input = split[0].rstrip()
            if input.startswith("http"):
                temp = input.split(" ")
                if len(temp) > 1:
                    input_type = "url"
                    i_tag1 = Tag.parse(["i", temp[0], input_type])
                    tags.append(i_tag1)
                    input_type = "text"
                    i_tag2 = Tag.parse(["i", input.replace(temp[0], "").lstrip(), input_type])
                    tags.append(i_tag2)
                else:
                    input_type = "url"
                    i_tag = Tag.parse(["i", input, input_type])
                    tags.append(i_tag)
            elif (input.startswith("nevent") or input.startswith("nostr:nevent") or input.startswith("note") or
                  input.startswith("nostr:note")):
                input_type = "event"
                if str(input).startswith('note'):
                    event_id = EventId.from_bech32(input)
                elif str(input).startswith("nevent"):
                    event_id = Nip19Event.from_bech32(input).event_id()
                elif str(input).startswith('nostr:note'):
                    event_id = EventId.from_nostr_uri(input)
                elif str(input).startswith("nostr:nevent"):
                    event_id = Nip19Event.from_nostr_uri(input).event_id()
                else:
                    event_id = EventId.from_hex(input)
                i_tag = Tag.parse(["i", event_id.to_hex(), input_type])
                tags.append(i_tag)
            else:
                print(input)
                input_type = "text"
                i_tag = Tag.parse(["i", input, input_type])
                tags.append(i_tag)

            alt_tag = Tag.parse(["alt", self.dvm_config.SUPPORTED_DVMS[index].TASK])
            tags.append(alt_tag)
            relaylist = ["relays"]
            for relay in self.dvm_config.RELAY_LIST:
                relaylist.append(relay)
            relays_tag = Tag.parse(relaylist)
            tags.append(relays_tag)
            output_tag = Tag.parse(["output", "text/plain"])
            tags.append(output_tag)
            remaining_text = command.replace(input, "")
            print(remaining_text)

            params = remaining_text.rstrip().split(" -")

            for i in params:
                print(i)
                if i != " ":
                    try:
                        split = i.split(" ")
                        if len(split) > 1:
                            param = str(split[0])
                            print(str(param))
                            value = str(split[1])
                            print(str(value))
                            if param == "cashu":
                                tag = Tag.parse([param, value])
                            else:
                                if param == "user":
                                    if value.startswith("@") or value.startswith("nostr:") or value.startswith("npub"):
                                        value = PublicKey.from_bech32(
                                            value.replace("@", "").replace("nostr:", "")).to_hex()
                                tag = Tag.parse(["param", param, value])
                            tags.append(tag)
                            print("Added params: " + str(tag.as_vec()))
                    except Exception as e:
                        print(e)
                        print("Couldn't add " + str(i))

            return tags

        def print_dvm_info(client, index):
            pubkey = self.dvm_config.SUPPORTED_DVMS[index].dvm_config.PUBLIC_KEY
            kind = self.dvm_config.SUPPORTED_DVMS[index].KIND
            nip89content_str = nip89_fetch_events_pubkey(client, pubkey, kind)
            print(nip89content_str)
            if nip89content_str is not None:
                nip89content = json.loads(nip89content_str)
                info = ""
                cashu_accepted = False
                encryption_supported = False

                if nip89content.get("name"):
                    info += "Name: " + nip89content.get("name") + "\n"
                if nip89content.get("image"):
                    info += nip89content.get("image") + "\n"
                if nip89content.get("about"):
                    info += "About:\n" + nip89content.get("about") + "\n\n"
                if nip89content.get("cashuAccepted"):
                    cashu_accepted = str(nip89content.get("cashuAccepted"))
                if nip89content.get("encryptionSupported"):
                    encryption_supported = str(nip89content.get("encryptionSupported"))

                info += "Encryption supported: " + str(encryption_supported) + "\n"
                info += "Cashu accepted: " + str(cashu_accepted) + "\n\n"
                if nip89content.get("nip90Params"):
                    params = nip89content["nip90Params"]
                    info += "\nParameters:\n"
                    for param in params:
                        info += "-" + param + '\n'
                        info += "Required: " + str(params[param]['required']) + '\n'
                        info += "Possible Values: " + json.dumps(params[param]['values']) + '\n\n'
                return info

            return None

        self.client.handle_notifications(NotificationHandler())

        try:
            while True:
                time.sleep(1.0)
        except KeyboardInterrupt:
            print('Stay weird!')
            os.kill(os.getpid(), signal.SIGTERM)



def build_lora_xl(lora, prompt, lora_weight):
    existing_lora = False
    if lora == "3drenderstyle":
        if lora_weight == "":
            lora_weight = "1"
        prompt = "3d style, 3d render, " + prompt + " <lora:3d_render_style_xl:"+lora_weight+">"
        existing_lora = True

    if lora == "psychedelicnoir":
        if lora_weight == "":
            lora_weight = "1"
        prompt = prompt + " <lora:Psychedelic_Noir__sdxl:"+lora_weight+">>"
        existing_lora = True

    if lora == "wojak":
        if lora_weight == "":
            lora_weight = "1"
        prompt = "<lora:wojak_big:"+lora_weight+">, " + prompt + ", wojak"
        existing_lora = True

    if lora == "dreamarts":
        if lora_weight == "":
            lora_weight = "1"
        prompt = "<lora:DreamARTSDXL:"+lora_weight+">, " + prompt
        existing_lora = True

    if lora == "voxel":
        if lora_weight == "":
            lora_weight = "1"
        prompt = "voxel style, " + prompt + " <lora:last:"+lora_weight+">"
        existing_lora = True

    if lora == "kru3ger":
        if lora_weight == "":
            lora_weight = "1"
        prompt = "kru3ger_style, " + prompt + "<lora:sebastiankrueger-kru3ger_style-000007:"+lora_weight+">"
        existing_lora = True

    if lora == "inkpunk":
        if lora_weight == "":
            lora_weight = "0.5"
        prompt = "inkpunk style, " + prompt + "  <lora:IPXL_v2:"+lora_weight+">"
        existing_lora = True

    if lora == "inkscenery":
        if lora_weight == "":
            lora_weight = "1"
        prompt = " ink scenery, " + prompt + " <lora:ink_scenery_xl:"+lora_weight+">"
        existing_lora = True

    if lora == "inkpainting":
        if lora_weight == "":
            lora_weight = "0.7"
        prompt = "painting style, " +  prompt + " <lora:Ink_Painting-000006::"+lora_weight+">,"
        existing_lora = True

    if lora == "timburton":
        if lora_weight == "":
            lora_weight = "1.27"
            pencil_weight = "1.15"
        prompt = prompt + " (hand drawn with pencil"+pencil_weight+"), (tim burton style:"+lora_weight+")" 
        existing_lora = True

    if lora == "pixelart":
        if lora_weight == "":
            lora_weight = "1"
        prompt = prompt + " (flat shading:1.2), (minimalist:1.4), <lora:pixelbuildings128-v2:"+lora_weight+"> "
        existing_lora = True

    if lora == "pepe":
        if lora_weight == "":
            lora_weight = "0.8"
        prompt = prompt + " ,<lora:DD-pepe-v2:"+lora_weight+"> pepe"
        existing_lora = True

    if lora == "bettertext":
        if lora_weight == "":
            lora_weight = "1"
        prompt = prompt + " ,<lora:BetterTextRedmond:"+lora_weight+">"
        existing_lora = True

    if lora == "mspaint":
        if lora_weight == "":
            lora_weight = "1"
        prompt = "MSPaint drawing " + prompt +">"
        existing_lora = True

    if lora == "woodfigure":
        if lora_weight == "":
            lora_weight = "0.7"
        prompt = prompt + ",woodfigurez,artistic style <lora:woodfigurez-sdxl:"+lora_weight+">"
        existing_lora = True

    if lora == "fireelement":
        prompt = prompt + ",composed of fire elements, fire element"
        existing_lora = True



    return lora, prompt, existing_lora


"""RealESRGan Module
"""

import os
import glob
import sys
from nova_utils.interfaces.server_module import Processor
from basicsr.archs.rrdbnet_arch import RRDBNet
from basicsr.utils.download_util import load_file_from_url
import numpy as np



from realesrgan import RealESRGANer
from realesrgan.archs.srvgg_arch import SRVGGNetCompact
import cv2
from PIL import Image as PILImage


# Setting defaults
_default_options = {"model": "RealESRGAN_x4plus", "outscale": 4, "denoise_strength": 0.5, "tile": 0,"tile_pad": 10,"pre_pad": 0, "compute_type": "fp32", "face_enhance": False }

# TODO: add log infos, 
class RealESRGan(Processor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.options = _default_options | self.options
        self.device = None
        self.ds_iter = None
        self.current_session = None
        self.model_path = None #Maybe need this later for manual path
       

        # IO shortcuts
        self.input = [x for x in self.model_io if x.io_type == "input"]
        self.output = [x for x in self.model_io if x.io_type == "output"]
        self.input = self.input[0]
        self.output = self.output[0]

    def process_data(self, ds_iter) -> dict:
        self.ds_iter = ds_iter
        current_session_name = self.ds_iter.session_names[0]
        self.current_session = self.ds_iter.sessions[current_session_name]['manager']
        input_image = self.current_session.input_data['input_image'].data
 

        try:
            model, netscale, file_url = self.manageModel(str(self.options['model']))

            if  self.model_path is not None:
                model_path = self.model_path
            else:
                model_path = os.path.join('weights', self.options['model'] + '.pth')
                if not os.path.isfile(model_path):
                    ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
                    for url in file_url:
                        # model_path will be updated
                        model_path = load_file_from_url(
                            url=url, model_dir=os.path.join(ROOT_DIR, 'weights'), progress=True, file_name=None)

             # use dni to control the denoise strength
            dni_weight = None
            if self.options['model'] == 'realesr-general-x4v3' and float(self.options['denoise_strength']) != 1:
                wdn_model_path = model_path.replace('realesr-general-x4v3', 'realesr-general-wdn-x4v3')
                model_path = [model_path, wdn_model_path]
                dni_weight = [float(self.options['denoise_strength']), 1 - float(self.options['denoise_strength'])]

            half = True
            if self.options["compute_type"] == "fp32":
                half=False


            upsampler = RealESRGANer(
            scale=netscale,
            model_path=model_path,
            dni_weight=dni_weight,
            model=model,
            tile= int(self.options['tile']),
            tile_pad=int(self.options['tile_pad']),
            pre_pad=int(self.options['pre_pad']),
            half=half,
            gpu_id=None) #Can be set if multiple gpus are available

            if bool(self.options['face_enhance']):  # Use GFPGAN for face enhancement
                from gfpgan import GFPGANer
                face_enhancer = GFPGANer(
                    model_path='https://github.com/TencentARC/GFPGAN/releases/download/v1.3.0/GFPGANv1.3.pth',
                    upscale=int(self.options['outscale']),
                    arch='clean',
                    channel_multiplier=2,
                    bg_upsampler=upsampler)
          
            
            pilimage =  PILImage.fromarray(input_image)
            img = cv2.cvtColor(np.array(pilimage), cv2.COLOR_RGB2BGR)
            try:
                if bool(self.options['face_enhance']):
                    _, _, output = face_enhancer.enhance(img, has_aligned=False, only_center_face=False, paste_back=True)
                else:
                    output, _ = upsampler.enhance(img, outscale=int(self.options['outscale']))
            except RuntimeError as error:
                print('Error', error)
                print('If you encounter CUDA out of memory, try to set --tile with a smaller number.')
    
            output = cv2.cvtColor(output, cv2.COLOR_BGR2RGB)

            return output
            
   


        except Exception as e:
            print(e)
            sys.stdout.flush()
            return "Error"

    
    def to_output(self, data: dict):
        self.current_session.output_data_templates['output_image'].data = data
        return self.current_session.output_data_templates


    def manageModel(self, model_name):
        if model_name == 'RealESRGAN_x4plus':  # x4 RRDBNet model
            model = RRDBNet(num_in_ch=3, num_out_ch=3, num_feat=64, num_block=23, num_grow_ch=32, scale=4)
            netscale = 4
            file_url = ['https://github.com/xinntao/Real-ESRGAN/releases/download/v0.1.0/RealESRGAN_x4plus.pth']
        elif model_name == 'RealESRNet_x4plus':  # x4 RRDBNet model
            model = RRDBNet(num_in_ch=3, num_out_ch=3, num_feat=64, num_block=23, num_grow_ch=32, scale=4)
            netscale = 4
            file_url = ['https://github.com/xinntao/Real-ESRGAN/releases/download/v0.1.1/RealESRNet_x4plus.pth']
        elif model_name == 'RealESRGAN_x4plus_anime_6B':  # x4 RRDBNet model with 6 blocks
            model = RRDBNet(num_in_ch=3, num_out_ch=3, num_feat=64, num_block=6, num_grow_ch=32, scale=4)
            netscale = 4
            file_url = ['https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.2.4/RealESRGAN_x4plus_anime_6B.pth']
        elif model_name == 'RealESRGAN_x2plus':  # x2 RRDBNet model
            model = RRDBNet(num_in_ch=3, num_out_ch=3, num_feat=64, num_block=23, num_grow_ch=32, scale=2)
            netscale = 2
            file_url = ['https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.1/RealESRGAN_x2plus.pth']
        elif model_name == 'realesr-animevideov3':  # x4 VGG-style model (XS size)
            model = SRVGGNetCompact(num_in_ch=3, num_out_ch=3, num_feat=64, num_conv=16, upscale=4, act_type='prelu')
            netscale = 4
            file_url = ['https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.5.0/realesr-animevideov3.pth']
        elif model_name == 'realesr-general-x4v3':  # x4 VGG-style model (S size)
            model = SRVGGNetCompact(num_in_ch=3, num_out_ch=3, num_feat=64, num_conv=32, upscale=4, act_type='prelu')
            netscale = 4
            file_url = [
                'https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.5.0/realesr-general-wdn-x4v3.pth',
                'https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.5.0/realesr-general-x4v3.pth'
            ]
        
        return model, netscale, file_url


# Copyright © 2023 Apple Inc.

import time
from typing import Tuple

import mlx.core as mx

from .model_io import (
    _DEFAULT_MODEL,
    load_autoencoder,
    load_diffusion_config,
    load_text_encoder,
    load_tokenizer,
    load_unet,
)
from .sampler import SimpleEulerSampler


def _repeat(x, n, axis):
    # Make the expanded shape
    s = x.shape
    s.insert(axis + 1, n)

    # Expand
    x = mx.broadcast_to(mx.expand_dims(x, axis + 1), s)

    # Make the flattened shape
    s.pop(axis + 1)
    s[axis] *= n

    return x.reshape(s)


class StableDiffusion:
    def __init__(self, model: str = _DEFAULT_MODEL, float16: bool = False):
        self.dtype = mx.float16 if float16 else mx.float32
        self.diffusion_config = load_diffusion_config(model)
        self.unet = load_unet(model, float16)
        self.text_encoder = load_text_encoder(model, float16)
        self.autoencoder = load_autoencoder(model, float16)
        self.sampler = SimpleEulerSampler(self.diffusion_config)
        self.tokenizer = load_tokenizer(model)

    def generate_latents(
        self,
        text: str,
        n_images: int = 1,
        num_steps: int = 50,
        cfg_weight: float = 7.5,
        negative_text: str = "",
        latent_size: Tuple[int] = (64, 64),
        seed=None,
    ):
        # Set the PRNG state
        seed = seed or int(time.time())
        mx.random.seed(seed)

        # Tokenize the text
        tokens = [self.tokenizer.tokenize(text)]
        if cfg_weight > 1:
            tokens += [self.tokenizer.tokenize(negative_text)]
        lengths = [len(t) for t in tokens]
        N = max(lengths)
        tokens = [t + [0] * (N - len(t)) for t in tokens]
        tokens = mx.array(tokens)

        # Compute the features
        conditioning = self.text_encoder(tokens)

        # Repeat the conditioning for each of the generated images
        if n_images > 1:
            conditioning = _repeat(conditioning, n_images, axis=0)

        # Create the latent variables
        x_T = self.sampler.sample_prior(
            (n_images, *latent_size, self.autoencoder.latent_channels), dtype=self.dtype
        )

        # Perform the denoising loop
        x_t = x_T
        for t, t_prev in self.sampler.timesteps(num_steps, dtype=self.dtype):
            x_t_unet = mx.concatenate([x_t] * 2, axis=0) if cfg_weight > 1 else x_t
            t_unet = mx.broadcast_to(t, [len(x_t_unet)])
            eps_pred = self.unet(x_t_unet, t_unet, encoder_x=conditioning)

            if cfg_weight > 1:
                eps_text, eps_neg = eps_pred.split(2)
                eps_pred = eps_neg + cfg_weight * (eps_text - eps_neg)

            x_t_prev = self.sampler.step(eps_pred, x_t, t, t_prev)
            x_t = x_t_prev
            yield x_t

    def decode(self, x_t):
        x = self.autoencoder.decode(x_t / self.autoencoder.scaling_factor)
        x = mx.minimum(1, mx.maximum(0, x / 2 + 0.5))
        return x


import json
import os
from datetime import timedelta
from threading import Thread

from nostr_sdk import Client, Timestamp, PublicKey, Tag, Keys, Options, SecretKey, NostrSigner, Kind, RelayOptions, \
    RelayLimits

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag
from nostr_dvm.utils.output_utils import post_process_list_to_users

"""
This File contains a Module to find inactive follows for a user on nostr

Accepted Inputs: None needed
Outputs: A list of users that have been inactive 
Params:  None
"""


class DiscoverInactiveFollows(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_PEOPLE_DISCOVERY
    TASK: str = "inactive-followings"
    FIX_COST: float = 100
    client: Client
    dvm_config: DVMConfig

    def __init__(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                 admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)
        super().__init__(name=name, dvm_config=dvm_config, nip89config=nip89config, nip88config=nip88config,
                         admin_config=admin_config, options=options)

    def is_input_supported(self, tags, client=None, dvm_config=None):
        # no input required
        return True

    def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        self.dvm_config = dvm_config

        request_form = {"jobID": event.id().to_hex()}

        # default values
        user = event.author().to_hex()
        since_days = 90

        for tag in event.tags():
            if tag.as_vec()[0] == 'param':
                param = tag.as_vec()[1]
                if param == "user":  # check for param type
                    user = tag.as_vec()[2]
                elif param == "since_days":  # check for param type
                    since_days = int(tag.as_vec()[2])

        options = {
            "user": user,
            "since_days": since_days
        }
        request_form['options'] = json.dumps(options)
        return request_form

    def process(self, request_form):
        from nostr_sdk import Filter
        from types import SimpleNamespace
        ns = SimpleNamespace()

        sk = SecretKey.from_hex(self.dvm_config.PRIVATE_KEY)
        keys = Keys.parse(sk.to_hex())
        signer = NostrSigner.keys(keys)

        #relaylimits = RelayLimits().event_max_num_tags(max_num_tags=10000)
        #relaylimits.event_max_size(None)
        relaylimits = RelayLimits.disable()

        opts = (Options().wait_for_send(False).send_timeout(timedelta(seconds=self.dvm_config.RELAY_TIMEOUT))).relay_limits(relaylimits)


        cli = Client.with_opts(signer, opts)
        for relay in self.dvm_config.RELAY_LIST:
            cli.add_relay(relay)
        ropts = RelayOptions().ping(False)
        cli.add_relay_with_opts("wss://nostr.band", ropts)

        cli.connect()

        options = self.set_options(request_form)
        step = 20

        followers_filter = Filter().author(PublicKey.parse(options["user"])).kind(Kind(3))
        followers = cli.get_events_of([followers_filter], timedelta(seconds=5))


        if len(followers) > 0:
            result_list = []
            newest = 0
            best_entry = followers[0]
            for entry in followers:
                print(len(best_entry.tags()))
                print(best_entry.created_at().as_secs())
                if entry.created_at().as_secs() > newest:
                    newest = entry.created_at().as_secs()
                    best_entry = entry


            print(best_entry.as_json())
            print(len(best_entry.tags()))
            print(best_entry.created_at().as_secs())
            print(Timestamp.now().as_secs())
            followings = []
            ns.dic = {}
            tagcount = 0
            for tag in best_entry.tags():
                tagcount += 1
                if tag.as_vec()[0] == "p":
                    following = tag.as_vec()[1]
                    followings.append(following)
                    ns.dic[following] = "False"
            print("Followings: " + str(len(followings)) + " Tags: " + str(tagcount))

            not_active_since_seconds = int(options["since_days"]) * 24 * 60 * 60
            dif = Timestamp.now().as_secs() - not_active_since_seconds
            not_active_since = Timestamp.from_secs(dif)

            def scanList(users, instance, i, st, notactivesince):
                from nostr_sdk import Filter

                keys = Keys.parse(self.dvm_config.PRIVATE_KEY)
                opts = Options().wait_for_send(True).send_timeout(
                    timedelta(seconds=5)).skip_disconnected_relays(True)
                signer = NostrSigner.keys(keys)
                cli = Client.with_opts(signer, opts)
                for relay in self.dvm_config.RELAY_LIST:
                    cli.add_relay(relay)
                cli.connect()

                filters = []
                for i in range(i, i + st):
                    filter1 = Filter().author(PublicKey.from_hex(users[i])).since(notactivesince).limit(1)
                    filters.append(filter1)
                event_from_authors = cli.get_events_of(filters, timedelta(seconds=10))
                for author in event_from_authors:
                    instance.dic[author.author().to_hex()] = "True"
                print(str(i) + "/" + str(len(users)))
                cli.disconnect()

            threads = []
            begin = 0
            # Spawn some threads to speed things up
            while begin < len(followings) - step:
                args = [followings, ns, begin, step, not_active_since]
                t = Thread(target=scanList, args=args)
                threads.append(t)
                begin = begin + step -1

            # last to step size
            missing_scans = (len(followings) - begin)
            args = [followings, ns, begin, missing_scans, not_active_since]
            t = Thread(target=scanList, args=args)
            threads.append(t)

            # Start all threads
            for x in threads:
                x.start()

            # Wait for all of them to finish
            for x in threads:
                x.join()

            result = {k for (k, v) in ns.dic.items() if v == "False"}

            print("Inactive accounts found: " + str(len(result)))
            for k in result:
                p_tag = Tag.parse(["p", k])
                result_list.append(p_tag.as_vec())

            return json.dumps(result_list)

    def post_process(self, result, event):
        """Overwrite the interface function to return a social client readable format, if requested"""
        for tag in event.tags():
            if tag.as_vec()[0] == 'output':
                format = tag.as_vec()[1]
                if format == "text/plain":  # check for output type
                    result = post_process_list_to_users(result)

        # if not text/plain, don't post-process
        return result


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    admin_config.LUD16 = dvm_config.LN_ADDRESS
    # Add NIP89
    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/c33ca6fc4cc038ca4adb46fdfdfda34951656f87ee364ef59095bae1495ce669.jpg",
        "about": "I discover users you follow, but that have been inactive on Nostr",
        "action": "unfollow", #follow, mute, unmute
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {
            "user": {
                "required": False,
                "values": [],
                "description": "Do the task for another user"
            },
            "since_days": {
                "required": False,
                "values": [],
                "description": "The number of days a user has not been active to be considered inactive"
            }
        }
    }
    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    return DiscoverInactiveFollows(name=name, dvm_config=dvm_config, nip89config=nip89config,
                                   admin_config=admin_config)


if __name__ == '__main__':
    process_venv(DiscoverInactiveFollows)



import os
from datetime import timedelta
from hashlib import sha256
from pathlib import Path

import dotenv
from nostr_sdk import Filter, Tag, Keys, EventBuilder, Client, EventId, PublicKey, Event, Timestamp, SingleLetterTag, \
    Alphabet, Kind

from nostr_dvm.utils import definitions
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.nostr_utils import send_event


class NIP88Config:
    DTAG: str = ""
    TITLE: str = ""
    CONTENT: str = ""
    IMAGE: str = ""
    TIER_EVENT: str = ""
    PERK1DESC: str = ""
    PERK2DESC: str = ""
    PERK3DESC: str = ""
    PERK4DESC: str = ""
    PAYMENT_VERIFIER_PUBKEY: str = ""

    AMOUNT_DAILY: int = None
    AMOUNT_MONTHLY: int = None
    AMOUNT_YEARLY: int = None


def nip88_create_d_tag(name, pubkey, image):
    key_str = str(name + image + pubkey)
    d_tag = sha256(key_str.encode('utf-8')).hexdigest()[:16]
    return d_tag


def fetch_nip88_parameters_for_deletion(keys, eventid, client, dvmconfig):
    idfilter = Filter().id(EventId.from_hex(eventid)).limit(1)
    nip88events = client.get_events_of([idfilter], timedelta(seconds=dvmconfig.RELAY_TIMEOUT))
    d_tag = ""
    if len(nip88events) == 0:
        print("Event not found. Potentially gone.")

    for event in nip88events:
        print(event.as_json())
        for tag in event.tags():
            if tag.as_vec()[0] == "d":
                d_tag = tag.as_vec()[1]
        if d_tag == "":
            print("No dtag found")
            return

        if event.author().to_hex() == keys.public_key().to_hex():
            nip88_delete_announcement(event.id().to_hex(), keys, d_tag, client, dvmconfig)
            print("NIP88 announcement deleted from known relays!")
        else:
            print("Privatekey does not belong to event")


def fetch_nip88_event(keys, eventid, client, dvmconfig):
    idfilter = Filter().id(EventId.parse(eventid)).limit(1)
    nip88events = client.get_events_of([idfilter], timedelta(seconds=dvmconfig.RELAY_TIMEOUT))
    d_tag = ""
    if len(nip88events) == 0:
        print("Event not found. Potentially gone.")

    for event in nip88events:

        for tag in event.tags():
            if tag.as_vec()[0] == "d":
                d_tag = tag.as_vec()[1]
        if d_tag == "":
            print("No dtag found")
            return

        if event.author().to_hex() == keys.public_key().to_hex():
            print(event.as_json())
        else:
            print("Privatekey does not belong to event")


def nip88_delete_announcement(eid: str, keys: Keys, dtag: str, client: Client, config):
    e_tag = Tag.parse(["e", eid])
    a_tag = Tag.parse(
        ["a", str(EventDefinitions.KIND_NIP88_TIER_EVENT) + ":" + keys.public_key().to_hex() + ":" + dtag])
    event = EventBuilder(Kind(5), "", [e_tag, a_tag]).to_event(keys)
    send_event(event, client, config)


def nip88_has_active_subscription(user: PublicKey, tiereventdtag, client: Client, receiver_public_key_hex, checkCanceled = True):
    subscription_status = {
        "isActive": False,
        "validUntil": 0,
        "subscriptionId": "",
        "expires": False,
    }

    subscriptionfilter = Filter().kind(definitions.EventDefinitions.KIND_NIP88_PAYMENT_RECIPE).pubkey(
        PublicKey.parse(receiver_public_key_hex)).custom_tag(SingleLetterTag.uppercase(Alphabet.P),
                                                             [user.to_hex()]).limit(1)
    evts = client.get_events_of([subscriptionfilter], timedelta(seconds=3))
    if len(evts) > 0:
        print(evts[0].as_json())
        matchesdtag = False
        for tag in evts[0].tags():
            if tag.as_vec()[0] == "valid":
                subscription_status["validUntil"] = int(tag.as_vec()[2])
            elif tag.as_vec()[0] == "e":
                subscription_status["subscriptionId"] = tag.as_vec()[1]
            elif tag.as_vec()[0] == "tier":
                if tag.as_vec()[1] == tiereventdtag:
                    matchesdtag = True

        if (subscription_status["validUntil"] > Timestamp.now().as_secs()) & matchesdtag:
            subscription_status["isActive"] = True

        if subscription_status["isActive"] and checkCanceled:
            # if subscription seems active, check if it has been canceled, and if so mark it as expiring.
            cancel_filter = Filter().kind(EventDefinitions.KIND_NIP88_STOP_SUBSCRIPTION_EVENT).author(
                user).pubkey(PublicKey.parse(receiver_public_key_hex)).event(
                EventId.parse(subscription_status["subscriptionId"])).limit(1)
            cancel_events = client.get_events_of([cancel_filter], timedelta(seconds=3))
            if len(cancel_events) > 0:
                if cancel_events[0].created_at().as_secs() > evts[0].created_at().as_secs():
                    subscription_status["expires"] = True

    return subscription_status


def nip88_announce_tier(dvm_config, client):
    title_tag = Tag.parse(["title", str(dvm_config.NIP88.TITLE)])
    image_tag = Tag.parse(["image", str(dvm_config.NIP88.IMAGE)])
    d_tag = Tag.parse(["d", dvm_config.NIP88.DTAG])

    # zap splits. Feel free to change this for your DVM

    # By default, 80% of subscription go to the current's DVM lightning address (make sure to update profile for it to work)
    # 5% go to NostrDVM developers
    # 5% go to NostrSDK developers
    # 10% optionally go to clients that support this subscription DVM
    zaptag1 = Tag.parse(["zap", dvm_config.PUBLIC_KEY, "wss://damus.io", "16"])
    zaptag2 = Tag.parse(["zap", "npub1nxa4tywfz9nqp7z9zp7nr7d4nchhclsf58lcqt5y782rmf2hefjquaa6q8", "wss://damus.io", "1"]) # NostrDVM
    zaptag3 = Tag.parse(["zap", "npub1drvpzev3syqt0kjrls50050uzf25gehpz9vgdw08hvex7e0vgfeq0eseet", "wss://damus.io", "1"]) # NostrSDK
    zaptag4 = Tag.parse(["zap", "", "wss://damus.io", "2"]) # Client might use this for splits
    p_tag = Tag.parse(["p", dvm_config.NIP88.PAYMENT_VERIFIER_PUBKEY])

    tags = [title_tag, image_tag, zaptag1, zaptag2, zaptag3, zaptag4, d_tag, p_tag]

    if dvm_config.NIP88.AMOUNT_DAILY is not None:
        amount_tag = Tag.parse(["amount", str(dvm_config.NIP88.AMOUNT_DAILY * 1000), "msats", "daily"])
        tags.append(amount_tag)

    if dvm_config.NIP88.AMOUNT_MONTHLY is not None:
        amount_tag = Tag.parse(["amount", str(dvm_config.NIP88.AMOUNT_MONTHLY * 1000), "msats", "monthly"])
        tags.append(amount_tag)

    if dvm_config.NIP88.AMOUNT_YEARLY is not None:
        amount_tag = Tag.parse(["amount", str(dvm_config.NIP88.AMOUNT_YEARLY * 1000), "msats", "yearly"])
        tags.append(amount_tag)

    if dvm_config.NIP88.PERK1DESC != "":
        perk_tag = Tag.parse(["perk", str(dvm_config.NIP88.PERK1DESC)])
        tags.append(perk_tag)
    if dvm_config.NIP88.PERK2DESC != "":
        perk_tag = Tag.parse(["perk", str(dvm_config.NIP88.PERK2DESC)])
        tags.append(perk_tag)
    if dvm_config.NIP88.PERK3DESC != "":
        perk_tag = Tag.parse(["perk", str(dvm_config.NIP88.PERK3DESC)])
        tags.append(perk_tag)
    if dvm_config.NIP88.PERK4DESC != "":
        perk_tag = Tag.parse(["perk", str(dvm_config.NIP88.PERK4DESC)])
        tags.append(perk_tag)

    keys = Keys.parse(dvm_config.NIP89.PK)
    content = dvm_config.NIP88.CONTENT
    event = EventBuilder(EventDefinitions.KIND_NIP88_TIER_EVENT, content, tags).to_event(keys)
    annotier_id = send_event(event, client=client, dvm_config=dvm_config)

    print("[" + dvm_config.NAME + "] Announced NIP 88 Tier for " + dvm_config.NIP89.NAME)

    return annotier_id

    # Relay and payment-verification


# ["r", "wss://my-subscribers-only-relay.com"],
# ["p", "<payment-verifier-pubkey>"],

def check_and_set_d_tag_nip88(identifier, name, pk, imageurl):
    if not os.getenv("NIP88_DTAG_" + identifier.upper()):
        new_dtag = nip88_create_d_tag(name, Keys.parse(pk).public_key().to_hex(),
                                      imageurl)
        nip88_add_dtag_to_env_file("NIP88_DTAG_" + identifier.upper(), new_dtag)
        print("Some new dtag:" + new_dtag)
        return new_dtag
    else:
        return os.getenv("NIP88_DTAG_" + identifier.upper())


def check_and_set_tiereventid_nip88(identifier, index="1", eventid=None):
    if eventid is None:
        if not os.getenv("NIP88_TIEREVENT_" + index + identifier.upper()):
            print("No Tier Event ID set")
            return None
        else:
            return os.getenv("NIP88_TIEREVENT_" + index + identifier.upper())
    else:
        nip88_add_dtag_to_env_file("NIP88_TIEREVENT_" + index + identifier.upper(), eventid)
        return eventid


def nip88_add_dtag_to_env_file(dtag, oskey):
    env_path = Path('.env')
    if env_path.is_file():
        print(f'loading environment from {env_path.resolve()}')
        dotenv.load_dotenv(env_path, verbose=True, override=True)
        dotenv.set_key(env_path, dtag, oskey)



# Copyright © 2023 Apple Inc.

import json
from functools import partial

import numpy as np
from huggingface_hub import hf_hub_download
from safetensors import safe_open as safetensor_open

import mlx.core as mx
from mlx.utils import tree_unflatten

from .clip import CLIPTextModel
from .config import UNetConfig, CLIPTextModelConfig, AutoencoderConfig, DiffusionConfig
from .tokenizer import Tokenizer
from .unet import UNetModel
from .vae import Autoencoder


_DEFAULT_MODEL = "stabilityai/stable-diffusion-2-1-base"
_MODELS = {
    # See https://huggingface.co/stabilityai/stable-diffusion-2-1-base for the model details and license
    "stabilityai/stable-diffusion-2-1-base": {
        "unet_config": "unet/config.json",
        "unet": "unet/diffusion_pytorch_model.safetensors",
        "text_encoder_config": "text_encoder/config.json",
        "text_encoder": "text_encoder/model.safetensors",
        "vae_config": "vae/config.json",
        "vae": "vae/diffusion_pytorch_model.safetensors",
        "diffusion_config": "scheduler/scheduler_config.json",
        "tokenizer_vocab": "tokenizer/vocab.json",
        "tokenizer_merges": "tokenizer/merges.txt",
    }
}


def _from_numpy(x):
    return mx.array(np.ascontiguousarray(x))


def map_unet_weights(key, value):
    # Map up/downsampling
    if "downsamplers" in key:
        key = key.replace("downsamplers.0.conv", "downsample")
    if "upsamplers" in key:
        key = key.replace("upsamplers.0.conv", "upsample")

    # Map the mid block
    if "mid_block.resnets.0" in key:
        key = key.replace("mid_block.resnets.0", "mid_blocks.0")
    if "mid_block.attentions.0" in key:
        key = key.replace("mid_block.attentions.0", "mid_blocks.1")
    if "mid_block.resnets.1" in key:
        key = key.replace("mid_block.resnets.1", "mid_blocks.2")

    # Map attention layers
    if "to_k" in key:
        key = key.replace("to_k", "key_proj")
    if "to_out.0" in key:
        key = key.replace("to_out.0", "out_proj")
    if "to_q" in key:
        key = key.replace("to_q", "query_proj")
    if "to_v" in key:
        key = key.replace("to_v", "value_proj")

    # Map transformer ffn
    if "ff.net.2" in key:
        key = key.replace("ff.net.2", "linear3")
    if "ff.net.0" in key:
        k1 = key.replace("ff.net.0.proj", "linear1")
        k2 = key.replace("ff.net.0.proj", "linear2")
        v1, v2 = np.split(value, 2)

        return [(k1, _from_numpy(v1)), (k2, _from_numpy(v2))]

    if "conv_shortcut.weight" in key:
        value = value.squeeze()

    # Transform the weights from 1x1 convs to linear
    if len(value.shape) == 4 and ("proj_in" in key or "proj_out" in key):
        value = value.squeeze()

    if len(value.shape) == 4:
        value = value.transpose(0, 2, 3, 1)

    return [(key, _from_numpy(value))]


def map_clip_text_encoder_weights(key, value):
    # Remove prefixes
    if key.startswith("text_model."):
        key = key[11:]
    if key.startswith("embeddings."):
        key = key[11:]
    if key.startswith("encoder."):
        key = key[8:]

    # Map attention layers
    if "self_attn." in key:
        key = key.replace("self_attn.", "attention.")
    if "q_proj." in key:
        key = key.replace("q_proj.", "query_proj.")
    if "k_proj." in key:
        key = key.replace("k_proj.", "key_proj.")
    if "v_proj." in key:
        key = key.replace("v_proj.", "value_proj.")

    # Map ffn layers
    if "mlp.fc1" in key:
        key = key.replace("mlp.fc1", "linear1")
    if "mlp.fc2" in key:
        key = key.replace("mlp.fc2", "linear2")

    return [(key, _from_numpy(value))]


def map_vae_weights(key, value):
    # Map up/downsampling
    if "downsamplers" in key:
        key = key.replace("downsamplers.0.conv", "downsample")
    if "upsamplers" in key:
        key = key.replace("upsamplers.0.conv", "upsample")

    # Map attention layers
    if "to_k" in key:
        key = key.replace("to_k", "key_proj")
    if "to_out.0" in key:
        key = key.replace("to_out.0", "out_proj")
    if "to_q" in key:
        key = key.replace("to_q", "query_proj")
    if "to_v" in key:
        key = key.replace("to_v", "value_proj")

    # Map the mid block
    if "mid_block.resnets.0" in key:
        key = key.replace("mid_block.resnets.0", "mid_blocks.0")
    if "mid_block.attentions.0" in key:
        key = key.replace("mid_block.attentions.0", "mid_blocks.1")
    if "mid_block.resnets.1" in key:
        key = key.replace("mid_block.resnets.1", "mid_blocks.2")

    # Map the quant/post_quant layers
    if "quant_conv" in key:
        key = key.replace("quant_conv", "quant_proj")
        value = value.squeeze()

    # Map the conv_shortcut to linear
    if "conv_shortcut.weight" in key:
        value = value.squeeze()

    if len(value.shape) == 4:
        value = value.transpose(0, 2, 3, 1)

    return [(key, _from_numpy(value))]


def _flatten(params):
    return [(k, v) for p in params for (k, v) in p]


def _load_safetensor_weights(mapper, model, weight_file, float16: bool = False):
    dtype = np.float16 if float16 else np.float32
    with safetensor_open(weight_file, framework="numpy") as f:
        weights = _flatten([mapper(k, f.get_tensor(k).astype(dtype)) for k in f.keys()])
    model.update(tree_unflatten(weights))


def _check_key(key: str, part: str):
    if key not in _MODELS:
        raise ValueError(
            f"[{part}] '{key}' model not found, choose one of {{{','.join(_MODELS.keys())}}}"
        )


def load_unet(key: str = _DEFAULT_MODEL, float16: bool = False):
    """Load the stable diffusion UNet from Hugging Face Hub."""
    _check_key(key, "load_unet")

    # Download the config and create the model
    unet_config = _MODELS[key]["unet_config"]
    with open(hf_hub_download(key, unet_config)) as f:
        config = json.load(f)

    n_blocks = len(config["block_out_channels"])
    model = UNetModel(
        UNetConfig(
            in_channels=config["in_channels"],
            out_channels=config["out_channels"],
            block_out_channels=config["block_out_channels"],
            layers_per_block=[config["layers_per_block"]] * n_blocks,
            num_attention_heads=[config["attention_head_dim"]] * n_blocks
            if isinstance(config["attention_head_dim"], int)
            else config["attention_head_dim"],
            cross_attention_dim=[config["cross_attention_dim"]] * n_blocks,
            norm_num_groups=config["norm_num_groups"],
        )
    )

    # Download the weights and map them into the model
    unet_weights = _MODELS[key]["unet"]
    weight_file = hf_hub_download(key, unet_weights)
    _load_safetensor_weights(map_unet_weights, model, weight_file, float16)

    return model


def load_text_encoder(key: str = _DEFAULT_MODEL, float16: bool = False):
    """Load the stable diffusion text encoder from Hugging Face Hub."""
    _check_key(key, "load_text_encoder")

    # Download the config and create the model
    text_encoder_config = _MODELS[key]["text_encoder_config"]
    with open(hf_hub_download(key, text_encoder_config)) as f:
        config = json.load(f)

    model = CLIPTextModel(
        CLIPTextModelConfig(
            num_layers=config["num_hidden_layers"],
            model_dims=config["hidden_size"],
            num_heads=config["num_attention_heads"],
            max_length=config["max_position_embeddings"],
            vocab_size=config["vocab_size"],
        )
    )

    # Download the weights and map them into the model
    text_encoder_weights = _MODELS[key]["text_encoder"]
    weight_file = hf_hub_download(key, text_encoder_weights)
    _load_safetensor_weights(map_clip_text_encoder_weights, model, weight_file, float16)

    return model


def load_autoencoder(key: str = _DEFAULT_MODEL, float16: bool = False):
    """Load the stable diffusion autoencoder from Hugging Face Hub."""
    _check_key(key, "load_autoencoder")

    # Download the config and create the model
    vae_config = _MODELS[key]["vae_config"]
    with open(hf_hub_download(key, vae_config)) as f:
        config = json.load(f)

    model = Autoencoder(
        AutoencoderConfig(
            in_channels=config["in_channels"],
            out_channels=config["out_channels"],
            latent_channels_out=2 * config["latent_channels"],
            latent_channels_in=config["latent_channels"],
            block_out_channels=config["block_out_channels"],
            layers_per_block=config["layers_per_block"],
            norm_num_groups=config["norm_num_groups"],
        )
    )

    # Download the weights and map them into the model
    vae_weights = _MODELS[key]["vae"]
    weight_file = hf_hub_download(key, vae_weights)
    _load_safetensor_weights(map_vae_weights, model, weight_file, float16)

    return model


def load_diffusion_config(key: str = _DEFAULT_MODEL):
    """Load the stable diffusion config from Hugging Face Hub."""
    _check_key(key, "load_diffusion_config")

    diffusion_config = _MODELS[key]["diffusion_config"]
    with open(hf_hub_download(key, diffusion_config)) as f:
        config = json.load(f)

    return DiffusionConfig(
        beta_start=config["beta_start"],
        beta_end=config["beta_end"],
        beta_schedule=config["beta_schedule"],
        num_train_steps=config["num_train_timesteps"],
    )


def load_tokenizer(key: str = _DEFAULT_MODEL):
    _check_key(key, "load_tokenizer")

    vocab_file = hf_hub_download(key, _MODELS[key]["tokenizer_vocab"])
    with open(vocab_file, encoding="utf-8") as f:
        vocab = json.load(f)

    merges_file = hf_hub_download(key, _MODELS[key]["tokenizer_merges"])
    with open(merges_file, encoding="utf-8") as f:
        bpe_merges = f.read().strip().split("\n")[1 : 49152 - 256 - 2 + 1]
    bpe_merges = [tuple(m.split()) for m in bpe_merges]
    bpe_ranks = dict(map(reversed, enumerate(bpe_merges)))

    return Tokenizer(bpe_ranks, vocab)
